#%%
import tensorflow as tf
import numpy as np
import logging
np.random.seed(42)

exp = tf.math.exp
pow = tf.math.pow
square = tf.math.square
sum = tf.math.reduce_sum
abs = tf.math.abs
mean = tf.math.reduce_mean
log = tf.math.log
sqrt = tf.math.sqrt

def inv(obj):
    return tf.math.reciprocal(obj + 1e-6)

def col_minmax(matrix, gene_id=None):
    if gene_id != None:
        if (np.max(matrix, axis=0) == np.min(matrix, axis=0)):
            print (gene_id)
            return matrix

    return (matrix - np.min(matrix, axis=0)) \
        / (np.max(matrix, axis=0) - np.min(matrix, axis=0))

def exp_args(adata, K=1):
    if adata.uns['base_function'] == 'Gaussian':
        columns = ['a', 'h', 'gamma', 'beta']
    else:
        columns = ['gamma', 'beta']
        columns.extend([f'a{k}' for k in range(K)])

    return columns

#%%
class Model_Utils():
    def __init__(
        self, 
        adata=None,
        var_names=None,
        Ms=None,
        Mu=None,
        config=None
    ):
        self.adata = adata
        self.var_names = var_names
        self.Ms, self.Mu = Ms, Mu
        self.K = 1
        self.win_size = 50
        self.config = config
        self.gene_log = []
        self.agenes_thres = int(config.AGENES_THRES * config.MAX_ITER)

    def init_vars(self):
        ngenes = self.Ms.shape[1]
        ones = tf.ones((1, ngenes), dtype=tf.float32)

        self.log_beta = tf.Variable(ones * 0, name='log_beta')
        self.intercept = tf.Variable(ones * 0, name='intercept')

        self.t = tf.Variable(ones * 0.5, name='t') #! mean of Gaussian
        self.log_a = tf.Variable(ones * 0, name='log_a') #! 1 / scaling of Gaussian
        self.offset = tf.Variable(ones * 0, name='offset')

        self.log_h = tf.Variable(ones * log(tf.math.reduce_max(self.Ms, axis=0)), name='log_h')

        init_gamma = self.adata.var['velocity_gamma'].values
        self.log_gamma = tf.Variable(
            log(tf.reshape(init_gamma, (1, self.adata.n_vars))), 
            name='log_gamma')

        for id in np.where(init_gamma <= 0)[0]:
            logging.info(f'name: {self.adata.var.index[id]}, gamma: {init_gamma[id]}')

        self.log_gamma = tf.Variable(
            tf.where(tf.math.is_finite(self.log_gamma), self.log_gamma, 0), 
            name='log_gamma')

        if self.config.VGENES == 'offset':
            init_inter = self.adata.var['velocity_inter'].values
            self.intercept = \
                tf.Variable(
                    tf.reshape(init_inter, (1, self.adata.n_vars)), 
                    name='intercept')

        if type(self.config.GENE_PRIOR) == list:
            vgenes_temp = []
            prior_t = np.ones((1, ngenes), dtype=np.float32) * 0.5
            for prior in self.config.GENE_PRIOR:
                prior_t[0][prior[2]] = 1.1 if prior[1] == 'increase' else -0.2
                vgenes_temp.append(prior + (prior_t[0][prior[2]], ))
            self.t = tf.Variable(prior_t, name='t')
            print (f'Modified GENE_PRIOR {vgenes_temp} with init_tau')

    def init_pars(self):
        self.default_pars_names = ['gamma', 'beta']
        self.default_pars_names += ['offset', 'a', 't', 'h', 'intercept']

    def init_weights(self, weighted=False):
        nonzero_s, nonzero_u = self.Ms > 0, self.Mu > 0
        weights = np.array(nonzero_s & nonzero_u, dtype=bool)

        if weighted:
            ub_s = np.percentile(self.s[weights], self.perc)
            ub_u = np.percentile(self.u[weights], self.perc)
            if ub_s > 0:
                weights &= np.ravel(self.s <= ub_s)
            if ub_u > 0:
                weights &= np.ravel(self.u <= ub_u)

        self.weights = tf.cast(weights, dtype=tf.float32)
        self.nobs = np.sum(weights, axis=0)

    def init_lr(self):
        if self.config.LEARNING_RATE == None:
            lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate=1e-2,
                decay_steps=2000, 
                decay_rate=0.9,
                staircase=True
            )

        else:
            lr_schedule = self.config.LEARNING_RATE

        return lr_schedule

    def get_fit_s(self, args, t_cell):
        self.fit_s = exp(args[5]) * \
            exp(-exp(args[3]) * square(t_cell - args[4])) + \
            args[2]

        return self.fit_s
    
    def get_s_deri(self, args, t_cell):
        self.s_deri = (self.fit_s - args[2]) * \
            (-exp(args[3]) * 2 * (t_cell - args[4]))

        return self.s_deri

    def get_fit_u(self, args):
        return (self.s_deri + exp(args[0]) * self.fit_s) / exp(args[1]) + args[6]
    
    def get_s_u(self, args, t_cell):
        s = self.get_fit_s(args, t_cell)
        s_deri = self.get_s_deri(args, t_cell)
        u = self.get_fit_u(args)

        if self.config.ASSIGN_POS_U:
            s = tf.clip_by_value(s, 0, 1000)
            u = tf.clip_by_value(u, 0, 1000)

        return s, u

    def max_density(self, dis):
        if self.config.DENSITY == 'Max':
            sort_list = np.sort(dis)
            diff = np.insert(np.diff(sort_list), 0, 0, axis=1)
            kernel = np.ones(self.win_size)

            loc = []
            for row in range(dis.shape[0]):
                idx = np.argmin(np.convolve(diff[row, :], kernel, 'valid'))
                window = sort_list[row, idx:idx + self.win_size]
                loc.append(np.mean(window))

            return np.array(loc, dtype=np.float32)

        if self.config.DENSITY == 'SVD':
            s, u, v = tf.linalg.svd(dis)
            s = s[0:50]
            u = u[:, :tf.size(s)]
            v = v[:tf.size(s), :tf.size(s)]
            dis_approx = tf.matmul(u, tf.matmul(tf.linalg.diag(s), v, adjoint_b=True))
            return tf.cast(mean(dis_approx, axis=1), tf.float32)

        if self.config.DENSITY == 'Raw':
            # weight = self.gene_prior_perc(dis)
            return tf.cast(mean(dis, axis=1), tf.float32)

    def gene_prior_perc(self, dis):
        perc, perc_idx = [], []
        aggregrate_weight = np.ones((dis.shape[1], ), dtype=np.float32)
    
        for prior in self.config.GENE_PRIOR:
            expr = np.array(self.adata[:, prior[0]].layers['Ms'])
            perc.append(np.max(expr) * 0.75) # modify 0.75 for parameter tuning
        
        perc_total = np.sum(perc)
        perc = [perc[i] / perc_total for i in range(len(perc))]

        for sequence, prior in enumerate(self.config.GENE_PRIOR):
            temp_id = np.sum(self.boolean[:prior[2]])
            aggregrate_weight[temp_id] = perc[sequence] * self.config.GENE_PRIOR_SCALE
            perc_idx.append(temp_id)

        # print (f'assigned weights of GENE_PRIOR {list(np.around(np.array(perc), 2)), perc_idx}')
        weight_total = np.sum(aggregrate_weight)
        aggregrate_weight = [aggregrate_weight[i] / weight_total for i in range(len(aggregrate_weight))]
        return np.reshape(aggregrate_weight, (1, dis.shape[1]))

    def match_time(self, Ms, Mu, s_predict, u_predict, x, iter):
        val = x[1, :] - x[0, :]
        cell_time = np.zeros((Ms.shape[0], Ms.shape[2]))

        if (self.config.AGENES_R2 < 1) & (iter > self.agenes_thres):
            self.index_list, self.boolean = np.squeeze(tf.where(self.total_genes)), self.total_genes
        else:
            self.index_list, self.boolean = np.squeeze(tf.where(self.idx)), self.idx

        for index in range(Ms.shape[2]):
            if index in self.index_list:
                sobs = Ms[:, :, index:index + 1] # n * 1 * 1
                uobs = Mu[:, :, index:index + 1]
                spre = s_predict[:, :, index:index + 1] # 1 * 3000 * 1
                upre = u_predict[:, :, index:index + 1]

                u_r2 = square(uobs - upre) # n * 3000 * 1
                s_r2 = square(sobs - spre)
                euclidean = sqrt(u_r2 + s_r2)
                assign_loc = tf.math.argmin(euclidean, axis=1).numpy() # n * 1
                
                if self.config.REORDER_CELL == 'Soft_Reorder':
                    cell_time[:, index:index + 1] = \
                        col_minmax(self.reorder(assign_loc), self.adata.var.index[index])
                if self.config.REORDER_CELL == 'Soft':
                    cell_time[:, index:index + 1] = \
                        col_minmax(assign_loc, self.adata.var.index[index])
                if self.config.REORDER_CELL == 'Hard':
                    cell_time[:, index:index + 1] = \
                        x[0, index:index + 1] + val[index:index + 1] * assign_loc

        if self.config.AGGREGATE_T:
            return self.max_density(cell_time[:, self.index_list]) #! sampling?
        else:
            return cell_time

    def reorder(self, loc):
        new_loc = np.zeros(loc.shape)

        for gid in range(loc.shape[1]):
            ref = sorted([(val, idx) for idx, val in enumerate(loc[:, gid])], 
                            key=lambda x:x[0])

            pre, count, rep = ref[0][0], 0, 0
            for item in ref:
                if item[0] > pre:
                    count += rep
                    rep = 1
                else:
                    rep += 1

                new_loc[item[1], gid] = count
                pre = item[0]
        
        return new_loc

    def init_time(self, boundary, shape=None):
        x = tf.linspace(boundary[0], boundary[1], shape[0])

        try:
            if type(boundary[0]) == int or boundary[0].shape[1] == 1:
                x = tf.reshape(x, (-1, 1))
                x = tf.broadcast_to(x, shape)
            else:
                x = tf.squeeze(x)
        except:
            x = tf.reshape(x, (-1, 1))
            x = tf.broadcast_to(x, shape)

        return tf.cast(x, dtype=tf.float32)
    
    def get_opt_args(self, iter, args):
        remain = iter % 400
        
        if self.config.FIT_OPTION == '1':
            if iter < self.config.MAX_ITER / 2:
                args_to_optimize = [args[2], args[3], args[4], args[5]] \
                    if remain < 200 else [args[0], args[1], args[6]]

            else:
                args_to_optimize = [args[0], args[1], 
                                    args[2], args[3],
                                    args[4], args[5], args[6]]
        
        if self.config.FIT_OPTION == '2':
            if iter < self.config.MAX_ITER / 2:
                args_to_optimize = [args[3], args[5]] \
                    if remain < 200 else [args[0], args[1]]
                    
            else:
                args_to_optimize = [args[0], args[1], args[3], args[5]]
        
        return args_to_optimize
    
    def get_log(self, loss, amplify=False, iter=None):
        self.finite = tf.math.is_finite(loss) # location of weird genes out of 2000
        glog = self.adata.var.iloc[np.squeeze(tf.where(~self.finite))].index.values 
        
        for gene in glog:
            if gene not in self.gene_log:
                logging.info(f'{gene}, iter {iter}')
                self.gene_log.append(gene)
            
            if amplify:
                if self.adata.var.at[gene, 'amplify_genes'] == True:
                    logging.info(f'{gene}, iter {iter}, amplify')
                    self.adata.var.at[gene, 'amplify_genes'] = False
        
        loss = tf.where(self.finite, loss, 0)
        return loss

    def get_loss(self, iter, s_r2, u_r2):
        if iter < self.config.MAX_ITER / 2:
            remain = iter % 400
            loss = s_r2 if remain < 200 else u_r2
        else:
            loss = s_r2 + u_r2 

        loss = self.get_log(loss, amplify=False, iter=iter)
        return loss
    
    def get_stop_cond(self, iter, pre, obj):
        stop_s = tf.zeros(self.Ms.shape[1], tf.bool)
        stop_u = tf.zeros(self.Ms.shape[1], tf.bool)
        remain = iter % 400

        if remain > 1 and remain < 200:
            stop_s = abs(pre - obj) <= abs(pre) * 1e-4
        if remain > 201 and remain < 400:
            stop_u = abs(pre - obj) <= abs(pre) * 1e-4

        return tf.math.logical_and(stop_s, stop_u)

    def get_interim_t(self, t_cell, idx):
        if self.config.AGGREGATE_T:
            return t_cell
        
        else:
            #? modify this later for independent mode
            t_interim = np.zeros(t_cell.shape)

            for i in range(t_cell.shape[1]):
                if idx[i]:
                    temp = np.reshape(t_cell[:, i], (-1, 1))
                    t_interim[:, i] = \
                        np.squeeze(col_minmax(temp, self.adata.var.index[i]))

            t_interim = self.max_density(t_interim)
            t_interim = tf.reshape(t_interim, (-1, 1))
            t_interim = tf.broadcast_to(t_interim, self.adata.shape)
            return t_interim


#%%
import matplotlib.pyplot as plt
import scvelo as scv
from IPython import display
import seaborn as sns
import numpy as np
import os
from .individual_gene import Validation

def plot_zero_gene_example(gene_name, adata, nonzero_para, zero_para):
    import math
    nrows = math.ceil(len(gene_name[:8]) / 2)

    fig, axes = plt.subplots(nrows, 2, figsize=(12, 5 * nrows))
    for i in range(nrows):
        sns.scatterplot(np.squeeze(adata[:, gene_name[i * 2]].layers['Ms']), 
                        np.squeeze(adata[:, gene_name[i * 2]].layers['Mu']), 
                        sizes=1, ax=axes[i][0], 
                        hue=adata.obs[adata.uns['label']], palette='tab20')
        axes[i][0].set_title(f'{gene_name[i * 2]}')

        nonzero_coef = np.round(nonzero_para.loc[gene_name[i * 2]]['coef'], 2)
        zero_coef = np.round(zero_para.loc[gene_name[i * 2]]['coef'], 2)
        nonzero_r2 = np.round(nonzero_para.loc[gene_name[i * 2]]['r2'], 2)
        zero_r2 = np.round(zero_para.loc[gene_name[i * 2]]['r2'], 2)

        axes[i][0].set_xlabel(f'({nonzero_coef}, {zero_coef})')
        axes[i][0].set_ylabel(f'({nonzero_r2}, {zero_r2})')
        
        sns.scatterplot(np.squeeze(adata[:, gene_name[i * 2 + 1]].layers['Ms']), 
                        np.squeeze(adata[:, gene_name[i * 2 + 1]].layers['Mu']), 
                        sizes=1, ax=axes[i][1], 
                        hue=adata.obs[adata.uns['label']], palette='tab20')
        axes[i][1].set_title(f'{gene_name[i * 2 + 1]}')

        nonzero_coef = np.round(nonzero_para.loc[gene_name[i * 2 + 1]]['coef'], 2)
        zero_coef = np.round(zero_para.loc[gene_name[i * 2 + 1]]['coef'], 2)
        nonzero_r2 = np.round(nonzero_para.loc[gene_name[i * 2 + 1]]['r2'], 2)
        zero_r2 = np.round(zero_para.loc[gene_name[i * 2 + 1]]['r2'], 2)

        axes[i][1].set_xlabel(f'({nonzero_coef}, {zero_coef})')
        axes[i][1].set_ylabel(f'({nonzero_r2}, {zero_r2})')

    plt.show()
    fig.savefig(os.path.join(adata.uns['temp'], 'Gene_Filter_ByZero.png'), dpi=300)
    from .pl import plot_zero_gene_distribution
    plot_zero_gene_distribution(nonzero_para, zero_para)

def plot_zero_gene_distribution(nonzero_para, zero_para, adata=None):
    fig, axes = plt.subplots(2, 2, figsize=(12, 10))
    sns.distplot(nonzero_para['coef'].values, ax=axes[0][0], kde=False)
    axes[0][0].set_title('Nonzero genes coefficient')

    sns.distplot(zero_para['coef'].values, ax=axes[0][1], kde=False)
    axes[0][1].set_title('Zero genes coefficient')

    sns.distplot(nonzero_para['r2'].values, ax=axes[1][0], kde=False)
    axes[1][0].set_title('Nonzero genes r2')

    sns.distplot(zero_para['r2'].values, ax=axes[1][1], kde=False)
    axes[1][1].set_title('Zero genes r2')
    plt.show()
    fig.savefig(os.path.join(adata.uns['temp'], 'Gene_Filter_ByZero_Distribution.png'), dpi=300)

def rbf(x, height, sigma, tau, offset_rbf):
    return height * np.exp(-sigma * (x - tau) * (x - tau)) + offset_rbf

def rbf_deri(x, height, sigma, tau, offset_rbf):
    return (rbf(x, height, sigma, tau, offset_rbf)  - offset_rbf) * (-sigma * 2 * (x - tau))

def rbf_u(x, height, sigma, tau, offset_rbf, beta, gamma, intercept):
    return (rbf_deri(x, height, sigma, tau, offset_rbf) + gamma * rbf(x, height, sigma, tau, offset_rbf)) / beta + intercept

def plot_range(
    gene_name, 
    adata, 
    config_file=None, 
    save_fig=False, 
    show_ax=False,
    show_legend=True,
    show_details=False,
    time_metric='latent_time',
    palette='tab20',
    size=20,
    ncols=1
):
    """
    Plotting function of phase portraits of individual genes.
    
    Args:
        gene_name (str): name of that gene to be illusrated, would extend to list of genes in next release
        adata (AnnData)
        config_file (.Config class): configuration file used for velocity estimation
        save_fig (bool): if True, save fig, default False
        
        show_ax (bool)
        show_legend (bool)
        show_details (bool): if True, plot detailed regression results together with estimated temporal change
        time_metric (str): inferred cell time, default 'latent_time'

        show_temporal (bool, experimental): whether plot temporal changes
            show_positive (bool, experimental): related to self.ASSIGN_POS_U
            t_left (float, experimental): starting time of phase portraits
            t_right (float, experimental): ending time of phase portraits
    """

    if config_file == None:
        raise ValueError('Please set attribute `config_file`')

    if time_metric == 'latent_time':
        if 'latent_time' not in adata.obs.columns:
            scv.tl.latent_time(adata, min_likelihood=None)

    if show_details:
        from .individual_gene import exam_genes        
        exam_genes(adata, gene_name, time_metric=time_metric)

    else:
        gene_name = gene_name if type(gene_name) == list else [gene_name]
        figs = []    

        for gn in gene_name:
            fig, axes = plt.subplots(
                    nrows=1,
                    ncols=3, 
                    figsize=(18, 4)
            )
            gdata = adata[:, gn]

            boundary = (gdata.var.fit_t.values - 3 * (1 / np.sqrt(2 * np.exp(gdata.var.fit_a.values))), 
                        gdata.var.fit_t.values + 3 * (1 / np.sqrt(2 * np.exp(gdata.var.fit_a.values))))
            
            t_one = np.linspace(0, 1, 1000)
            t_boundary = np.linspace(boundary[0], boundary[1], 2000)

            spre = np.squeeze(rbf(t_boundary, gdata.var.fit_h.values, gdata.var.fit_a.values, gdata.var.fit_t.values, gdata.var.fit_offset.values))
            sone = np.squeeze(rbf(t_one, gdata.var.fit_h.values, gdata.var.fit_a.values, gdata.var.fit_t.values, gdata.var.fit_offset.values))

            upre = np.squeeze(rbf_u(t_boundary, gdata.var.fit_h.values, gdata.var.fit_a.values, gdata.var.fit_t.values, gdata.var.fit_offset.values, gdata.var.fit_beta.values, gdata.var.fit_gamma.values, gdata.var.fit_intercept.values))
            uone = np.squeeze(rbf_u(t_one, gdata.var.fit_h.values, gdata.var.fit_a.values, gdata.var.fit_t.values, gdata.var.fit_offset.values, gdata.var.fit_beta.values, gdata.var.fit_gamma.values, gdata.var.fit_intercept.values))

            g1 = sns.scatterplot(x=np.squeeze(gdata.layers['Ms']), 
                                y=np.squeeze(gdata.layers['Mu']), 
                                s=size, hue=adata.obs[adata.uns['label']], 
                                palette=palette, ax=axes[0])
            axes[0].plot(spre, upre, color='lightgrey', linewidth=2, label='Predicted Curve')
            axes[0].plot(sone, uone, color='black', linewidth=2, label='Predicted Curve Time 0-1')
            axes[0].set_xlabel('Spliced Reads')
            axes[0].set_ylabel('Unspliced Reads')

            axes[0].set_xlim([-0.005 if gdata.layers['Ms'].min() < 1
                    else gdata.layers['Ms'].min() * 0.95, 
                    gdata.layers['Ms'].max() * 1.05])
            axes[0].set_ylim([-0.005 if gdata.layers['Mu'].min() < 1
                    else gdata.layers['Mu'].min() * 0.95, 
                    gdata.layers['Mu'].max() * 1.05])

            g2 = sns.scatterplot(x=np.squeeze(adata.obs[time_metric].values), 
                                y=np.squeeze(gdata.layers['Ms']), 
                                s=size, hue=adata.obs[adata.uns['label']], 
                                palette=palette, ax=axes[1])
            sns.lineplot(x=t_one, y=sone, color='black', linewidth=2, ax=axes[1])

            axes[1].set_xlabel('Inferred Cell Time')
            axes[1].set_ylabel('Spliced')

            g3 = sns.scatterplot(x=np.squeeze(adata.obs[time_metric].values), 
                                y=np.squeeze(gdata.layers['Mu']), 
                                s=size, hue=adata.obs[adata.uns['label']], 
                                palette=palette, ax=axes[2])
            sns.lineplot(x=t_one, y=uone, color='black', linewidth=2, ax=axes[2])

            axes[2].set_xlabel('Inferred Cell Time')
            axes[2].set_ylabel('Unspliced')

            # if not show_ax:
            #     axes.axis("off")

            if not show_legend:
                g1.get_legend().remove()
                g2.get_legend().remove()
                g3.get_legend().remove()

            axes[1].set_title(gn, fontsize=12)
            plt.show()

            if save_fig:
                plt.savefig(os.path.join(adata.uns['temp'], f'GM_{gn}.png'), dpi=300, bbox_inches='tight')

def plot_phase_portrait(adata, args, sobs, uobs, spre, upre):
    if 'examine_genes' in adata.uns.keys():
        display.clear_output(wait=True)
        examine = Validation(adata)
        fig, axes = plt.subplots(1, 1, figsize=(6, 4))

        examine.plot_mf(adata, sobs, uobs, axes[0])
        examine.plot_mf(adata, spre, upre, axes[0])
        plt.show()
    
    else:
        pass

def plot_cell_time(adata):
    if 'examine_genes' in adata.uns.keys():
        raise ValueError(
            f'self.VGENES in configuration file should not be a specified gene name.\n'
            f'Please re-run the model use alternative setting.'
        )
    
    else:
        scv.pl.scatter(
            adata, color='latent_time', cmap='gnuplot', 
            size=25, title='Assigned cell time', dpi=300
        )

def plot_loss(iter, loss, thres=None):
    fig, axes = plt.subplots(1, 2, figsize=(12, 4))
    x = range(iter + 1)

    subiter, subloss = x[800:thres - 1], loss[800:thres - 1]
    axes[0].plot(subiter, subloss)
    axes[0].set_title('Iter # from 800 to cutoff')
    axes[0].set_ylabel('Euclidean Loss')

    # subiter, subloss = x[int(iter / 2):], loss[int(iter / 2):]
    # axes[1].plot(subiter, subloss)
    # axes[1].set_title('Iter # from 1/2 of maximum')

    subiter, subloss = x[int(thres * 1.01):], loss[int(thres * 1.01):]
    axes[1].plot(subiter, subloss)
    axes[1].set_title('Iter # from cutoff to terminated state')

    plt.show()
    plt.close()

def plot_compare_loss(adata):
    var = adata.var.loc[adata.var['velocity_genes'] == True]

    fig, axes = plt.subplots(1, 2, figsize=(14, 4))
    x, y = np.log(var['li_loss']), np.log(var['fit_loss'])
    sns.scatterplot(x, y, ax=axes[0])
    axes[0].plot([np.min(x), np.max(x)], [np.min(x), np.max(x)], linestyle='--', c='r')
    axes[0].set_title('Log Loss')

    x, y = var['li_loss'], var['fit_loss']
    sns.scatterplot(x, y, ax=axes[1])
    axes[1].plot([np.min(x), np.max(x)], [np.min(x), np.max(x)], linestyle='--', c='r')
    axes[1].set_title('Normal Loss')

    plt.show()
    plt.close()

    print (f'---> # of genes which linear regression loss is smaller: ', end='')
    print (var.loc[var['fit_loss'] >= var['li_loss']].shape[0])

def plot_compare_bic(adata):
    var = adata.var.loc[adata.var['velocity_genes'] == True]

    fig, axes = plt.subplots(1, 1, figsize=(6, 4))
    x, y = var['li_bic'], var['fit_bic']
    sns.scatterplot(x, y, ax=axes)
    axes.plot([np.min(x), np.max(x)], [np.min(x), np.max(x)], linestyle='--', c='r')
    axes.set_title('Normal BIC')

    plt.show()
    plt.close()

    print (f'---> # of genes which linear regression BIC is smaller: ', end='')
    print (var.loc[var['fit_bic'] >= var['li_bic']].shape[0])
    print (var.loc[var['fit_bic'] >= var['li_bic']].shape[0] / var.shape[0])

def plot_compare_llf(adata):
    var = adata.var.loc[adata.var['velocity_genes'] == True]

    fig, axes = plt.subplots(1, 1, figsize=(6, 4))
    ratio = -2 * (var['li_llf'] - var['fit_llf'])
    sns.distplot(ratio, ax=axes, bins=200, kde=True)
    axes.set_title('Likelihood Ratio')
    
    plt.show()
    plt.close()

def plot_reverse_tran_scatter(adata):
    sns.scatterplot(x='rbf_r2', y='qua_r2', 
        data=adata.var.loc[adata.var['velocity_genes'] == True])
    plt.axline((0, 0), (0.5, 0.5), color='r')
        
    plt.title(f'$R^2$ comparison of RBF and Quadratic model')
    plt.show()
    plt.close()


#%%
import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix, csr_matrix

def vals_to_csr(vals, rows, cols, shape):
    graph = coo_matrix((vals, (rows, cols)), shape=shape)
    graph_neg = graph.copy()

    graph.data = np.clip(graph.data, 0, 1)
    graph_neg.data = np.clip(graph_neg.data, -1, 0)

    graph.eliminate_zeros()
    graph_neg.eliminate_zeros()

    return graph.tocsr(), graph_neg.tocsr()

def get_iterative_indices(indices, index, n_recurse_neighbors=2):
    def iterate_indices(indices, index, n_recurse_neighbors):
        if n_recurse_neighbors > 1:
            index = iterate_indices(indices, index, n_recurse_neighbors - 1)
        ix = np.append(index, indices[index])  # direct and indirect neighbors
        if np.isnan(ix).any():
            ix = ix[~np.isnan(ix)]
        return ix.astype(int)

    indices = np.unique(iterate_indices(indices, index, n_recurse_neighbors))
    return indices

def cosine_correlation(dX, Vi):
    dx = dX - dX.mean(-1)[:, None]
    Vi_norm = vector_norm(Vi)

    if Vi_norm == 0:
        result = np.zeros(dx.shape[0])
    else:
        result = np.einsum("ij, j", dx, Vi) / (norm(dx) * Vi_norm)[None, :]
        
    return result

def vector_norm(x):
    """computes the L2-norm along axis 1, equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(np.einsum("i, i -> ", x, x))

def norm(A):
    """computes the L2-norm along axis 1, equivalent to np.linalg.norm(A, axis=1)
    """
    return np.sqrt(np.einsum("ij, ij -> i", A, A) if A.ndim > 1 else np.sum(A * A))

class Influence():
    def __init__(self) -> None:
        pass

    def get_n_jobs(self, n_jobs):
        import os

        if n_jobs is None or (n_jobs < 0 and os.cpu_count() + 1 + n_jobs <= 0):
            return 1, os.cpu_count()
        elif n_jobs > os.cpu_count():
            return os.cpu_count(), os.cpu_count()
        elif n_jobs < 0:
            return os.cpu_count() + 1 + n_jobs, os.cpu_count()
        else:
            return n_jobs, os.cpu_count()

    def get_indices(self, D):
        D.data += 1e-6
        A = D > 0
        n_counts = A.sum(1).A1 # summation over axis 1, equivalent to np.sum(A, 1)

        D.eliminate_zeros()
        D.data -= 1e-6
        indices = D.indices.reshape((-1, n_counts.min()))
        
        return indices, D

    def compute_cosines(self):
        vals, rows, cols = [], [], []

        for i in range(self.X.shape[0]):
            if self.V[i].max() != 0 or self.V[i].min() != 0:
                neighs_idx = get_iterative_indices(self.indices, i)

                dX = self.X[neighs_idx] - self.X[i, None]  # 60% of runtime
                dX = np.sqrt(np.abs(dX)) * np.sign(dX)
                val = cosine_correlation(dX, self.V[i])  # 40% of runtime

                vals.extend(val)
                rows.extend(np.ones(len(neighs_idx)) * i)
                cols.extend(neighs_idx)

        vals = np.hstack(vals)
        vals[np.isnan(vals)] = 0
        
        graph, graph_neg = \
            vals_to_csr(vals, rows, cols, shape=(self.X.shape[0], self.X.shape[0]))
        return graph, graph_neg

    def transition_matrix(self, gene_list):
        #! vgraph = VelocityGraph(adata, sqrt_transform=True, gene_subset=gene_list)
        used_genes = self.var_names.isin(gene_list)
        self.X = np.array(self.Ms[:, used_genes], dtype=np.float32)
        self.V = np.array(self.velocity[:, used_genes], dtype=np.float32)

        self.V = np.sqrt(np.abs(self.V)) * np.sign(self.V)
        self.V -= np.nanmean(self.V, axis=1)[:, None]

        self.indices = self.get_indices(self.distance)[0]

        #! vgraph.compute_cosines()
        graph, graph_neg = self.compute_cosines()

        #! transition matrix in sparse format
        graph = csr_matrix(graph)

        confidence = graph.max(1).A.flatten()
        ub = np.percentile(confidence, 98)
        self_prob = np.clip(ub - confidence, 0, 1)
        graph.setdiag(self_prob)

        T = np.expm1(graph * 10)  # np.exp(graph.A * 10) - 1
        T -= np.expm1(-graph_neg * 10)

        T = T.multiply(csr_matrix(1.0 / np.abs(T).sum(1))) # original `normalize` function
        T.setdiag(0)
        T.eliminate_zeros()
        return T.A

    def get_importance_simplify(self, args):
        gene_list, gene = args
        T = self.transition_matrix(gene_list)

        #! self.get_importance by clusters or entire dataset
        cosine = []
        cosine.append(np.abs(self.Tref - T).sum())

        for type in self.ctypes:
            index = np.squeeze(np.argwhere(self.label_list == type))
            cosine.append(np.abs(self.Tref[index, :] - T[index, :]).sum())
        
        return [cosine, gene]

    def verify_neighbors(self, adata):
        valid = "neighbors" in adata.uns.keys() and "params" in adata.uns["neighbors"]
        
        if valid:
            n_neighs = (adata.obsp["distances"] > 0).sum(1)
            valid = n_neighs.min() * 2 > n_neighs.max()

        if not valid:
            raise ValueError("You need to run scv.pp.neighbors first.")

    def recover_importance(
        self,
        adata,
        n_jobs=None
    ):
        """Rank genes importance (influence) on estimated velocity fields
            by filtering out one gene by another and compare the cosine similarity
            between original velocity embedding (equivalent to transition matrix)
            with the new velocity embedding minus one genes
        
        Args:
            adata (Anndata): 
                Anndata object.
            n_jobs (int): 
                number of CPU cores to use.
            
        Returns:
            dataframe: 
                calculated scores for each gene (unranked version)
                need to run `self.rank_importance(df)`
        
        """

        import time
        import multiprocessing
        self.verify_neighbors(adata)

        basis = adata.uns['basis']

        if f"X_{basis}" not in adata.obsm_keys():
            raise ValueError("You need to compute the embedding first.")

        if f'velocity_graph' not in adata.uns.keys():
            raise ValueError('Need to run `scv.tl.velocity_graph(adata, gene_subset=None)` first')

        vgenes = adata.var.loc[adata.var['velocity_genes'] == True].index
        n_jobs, total_jobs = self.get_n_jobs(n_jobs=n_jobs)

        self.adata = adata
        self.var_names = adata.var_names
        self.Ms = adata.layers["Ms"]
        self.velocity = adata.layers["velocity"]
        self.distance = adata.obsp["distances"]
        self.label_list = adata.obs[adata.uns['label']].values
        self.ctypes = list(set(adata.obs[adata.uns['label']].values))
        self.Tref = self.transition_matrix(list(vgenes))

        ctime = time.time()
        print (f"(using {n_jobs}/{total_jobs} cores)")
        with multiprocessing.Pool(n_jobs) as pool:
            self.res = pool.map_async(
                self.get_importance_simplify, 
                self.bufferize(vgenes)).get()
        print ('Time elapsed ', int(time.time() - ctime), ' seconds.')

        columns = ['Overall']
        columns.extend([str(type) for type in self.ctypes])
        df_aggre = pd.DataFrame(index=vgenes, data=0, dtype=np.float32, columns=columns)

        for _res in self.res:
            df_aggre.at[_res[1], :] = _res[0] 

        return (df_aggre - df_aggre.min(axis=0)) / \
            (df_aggre.max(axis=0) - df_aggre.min(axis=0))

    def bufferize(self, velocity_genes):
        buffer = []

        for gene in velocity_genes:
            used_genes = list(velocity_genes).copy()
            used_genes.remove(gene)

            buffer.append([used_genes, gene])
        
        return buffer

    def rank_importance(self, df):
        ranking = pd.DataFrame(index=list(range(len(df))), data=np.nan, 
            dtype=np.float32, columns=df.columns)
        
        for col in ranking.columns:
            ranking.at[:, col] = df.sort_values(by=[col], ascending=False).index
        
        return ranking
    
def influence(adata, n_jobs):
    import os
    from .gene_influence import Influence
    val = Influence()

    gene_score = val.recover_importance(adata, n_jobs)
    ranking = val.rank_importance(gene_score)

    gene_score = gene_score.sort_values(by=['Overall'], ascending=False)
    adata.uns['gene_rank'] = ranking
    adata.uns['gene_score'] = gene_score

    adata.write(os.path.join(adata.uns['temp'], 'temp.h5ad'), compression='gzip')

    return ranking, gene_score


import numpy as np
import os
import pandas as pd
from .model import lagrange
from .utils import make_dense, get_weight, R2
import scvelo as scv
import tensorflow as tf

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
np.random.seed(42)

class Velocity:
    def __init__(
        self,
        adata=None,
        min_ratio=0.01,
        min_r2=0.01,
        fit_offset=False,
        perc=[5, 95],
        vkey='velocity',
        config=None
    ):
        self.adata = adata
        self.vkey = vkey

        self.Ms = adata.layers["spliced"] if config.USE_RAW else adata.layers["Ms"].copy()
        self.Mu = adata.layers["unspliced"] if config.USE_RAW else adata.layers["Mu"].copy()
        self.Ms, self.Mu = make_dense(self.Ms), make_dense(self.Mu)

        self.min_r2 = min_r2
        self.min_ratio = min_ratio
        
        n_obs, n_vars = self.Ms.shape
        self.gamma = np.zeros(n_vars, dtype=np.float32)
        self.r2 = np.zeros(n_vars, dtype=np.float32)
        self.velocity_genes = np.ones(n_vars, dtype=bool)
        self.residual_scale = np.zeros([n_obs, n_vars], dtype=np.float32)
    
        self.perc = perc
        self.fit_offset = fit_offset
        self.config = config

    def get_velo_genes(self):
        variable = self.adata.var
        if variable.index[0].startswith('ENSMUSG'):
            variable.index = variable['gene']
            variable.index.name = 'index' 
        
        weights = get_weight(self.Ms, self.Mu, perc=95)
        Ms, Mu = weights * self.Ms, weights * self.Mu

        self.gamma_quantile = np.sum(Mu * Ms, axis=0) / np.sum(Ms * Ms, axis=0)
        self.scaling = np.std(self.Mu, axis=0) / np.std(self.Ms, axis=0)
        self.adata.layers['Mu_scale'] = self.Mu / self.scaling

        if self.config.R2_ADJUST:
            Ms, Mu = self.Ms, self.Mu

        self.gene_index = variable.index
        self.gamma_ref = np.sum(Mu * Ms, axis=0) / np.sum(Ms * Ms, axis=0)
        self.residual_scale = self.Mu - self.gamma_ref * self.Ms
        self.r2 = R2(self.residual_scale, total=self.Mu - np.mean(self.Mu, axis=0))

        self.velocity_genes = np.ones(Ms.shape[1], dtype=bool)

        if type(self.config.VGENES) == str and \
            self.config.VGENES in self.adata.var.index:
            self.adata.uns['examine_genes'] = self.config.VGENES
            self.velocity_genes = np.zeros(Ms.shape[1], dtype=np.bool)
            self.velocity_genes[
                np.argwhere(self.adata.var.index == self.config.VGENES)] = True

        elif type(self.config.VGENES) == list:
            temp = []
            for gene in variable.index:
                if gene in self.config.VGENES:
                    temp.append(True)
                else:
                    temp.append(False)
                    
            self.velocity_genes = np.array(temp)
            self.adata.var['velocity_gamma'] = self.gamma_ref

        elif self.config.VGENES == 'raws':
            self.velocity_genes = np.ones(Ms.shape[1], dtype=np.bool)

        elif self.config.VGENES == 'offset':
            self.fit_linear(None, self.Ms, self.Mu, 'vgene_offset', coarse=False)

            if self.config.FILTER_CELLS:
                self.fit_linear(None, self.Ms, self.Mu, 'vgene_offset', coarse=True)

        elif self.config.VGENES == 'basic':
            self.velocity_genes = (
                (self.r2 > self.min_r2)
                & (self.r2 < 0.95)
                & (self.gamma_quantile > self.min_ratio)
                & (self.gamma_ref > self.min_ratio)
                & (np.max(self.Ms > 0, axis=0) > 0)
                & (np.max(self.Mu > 0, axis=0) > 0)
            )
            print (f'# of velocity genes {self.velocity_genes.sum()} (Criterion: positive regression coefficient between un/spliced counts)')
            
            if self.config.R2_ADJUST:
                lb, ub = np.nanpercentile(self.scaling, [10, 90])
                self.velocity_genes = (
                    self.velocity_genes
                    & (self.scaling > np.min([lb, 0.03]))
                    & (self.scaling < np.max([ub, 3]))
                )
            print (f'# of velocity genes {self.velocity_genes.sum()} (Criterion: std of un/spliced reads should be moderate, w/o extreme values)')

            self.adata.var['velocity_gamma'] = self.gamma_ref
            self.adata.var['velocity_r2'] = self.r2

        else:
            raise ValueError('Plase specify the correct self.VGENES in configuration file')

        if True:
            self.init_weights()
            self.velocity_genes = self.velocity_genes & (self.nobs > 0.05 * Ms.shape[1])

        self.adata.var['scaling'] = self.scaling
        self.adata.var['velocity_genes'] = self.velocity_genes
        self.adata.uns[f"{self.vkey}_params"] = {"mode": self.config.GENERAL, "perc": self.perc}
        
        if np.sum(self.velocity_genes) < 2:
            print ('---> Low signal in splicing dynamics.')

        from .utils import prior_trend_valid
        if type(self.config.IROOT) == list:
            self.config.IROOT = prior_trend_valid(self.adata, self.config.IROOT, 'IROOT')
        
        if type(self.config.GENE_PRIOR) == list:
            self.config.GENE_PRIOR = prior_trend_valid(self.adata, self.config.GENE_PRIOR, 'GENE_PRIOR')

    def init_weights(self):
        nonzero_s, nonzero_u = self.Ms > 0, self.Mu > 0
        weights = np.array(nonzero_s & nonzero_u, dtype=bool)
        self.nobs = np.sum(weights, axis=0)

    def fit_deterministic(self, idx, Ms, Mu, Ms_scale, Mu_scale):
        df_gamma = pd.DataFrame(index=self.gene_index, data=0, 
            dtype=np.float32, columns=['coef', 'inter'])

        if self.fit_offset:
            pass

        else:
            weights_new = get_weight(Ms, Mu, perc=self.perc)
            x, y = weights_new * Ms_scale, weights_new * Mu_scale
            df_gamma['coef'][idx] = np.sum(y * x, axis=0) / np.sum(x * x, axis=0)

        residual = self.Mu \
            - np.broadcast_to(df_gamma['coef'].values, self.Mu.shape) * self.Ms \
            - np.broadcast_to(df_gamma['inter'].values, self.Mu.shape)
        
        self.adata.var['velocity_gamma'] = df_gamma['coef'].values
        self.adata.var['intercept'] = df_gamma['inter'].values
        return residual

    def fit_linear(self, idx, Ms, Mu, method='vgene_offset', coarse=False):
        '''
        [bic] for linear BIC comparison with algorithm BIC
        [kinetic] for selection of cyclic kinetic gene and monotonic expression gene
        [vgene_offset] for determination of velocity gene alternatively using offset
        '''

        if method == 'vgene_offset':
            from sklearn import linear_model
            from sklearn.metrics import r2_score

            index = self.adata.var.index
            linear = pd.DataFrame(index=index, data=0, dtype=np.float32, 
                columns=['coef', 'inter', 'r2'])

            reg = linear_model.LinearRegression()
            for col in range(Ms.shape[1]):
                if not coarse:
                    sobs = np.reshape(Ms[:, col], (-1, 1))
                    uobs = np.reshape(Mu[:, col], (-1, 1))

                else:                
                    if self.config.FILTER_CELLS:
                        nonzero_s = Ms[:, col] > 0
                        nonzero_u = Mu[:, col] > 0
                        valid = np.array(nonzero_s & nonzero_u, dtype=bool)
                        sobs = np.reshape(Ms[:, col][valid], (-1, 1))
                        uobs = np.reshape(Mu[:, col][valid], (-1, 1))

                    else:
                        sobs = np.reshape(Ms[:, col], (-1, 1))
                        uobs = np.reshape(Mu[:, col], (-1, 1))

                reg.fit(sobs, uobs)
                u_pred = reg.predict(sobs)

                linear.loc[index[col], 'coef'] = float(reg.coef_)
                linear.loc[index[col], 'inter'] = float(reg.intercept_)
                linear.loc[index[col], 'r2'] = r2_score(uobs, u_pred)

            self.adata.var['velocity_inter'] = np.array(linear['inter'].values)
            self.adata.var['velocity_gamma'] = np.array(linear['coef'].values)
            self.adata.var['velocity_r2'] = np.array(linear['r2'].values)
            self.gamma_ref = np.array(linear['coef'].values)
            self.r2 = np.array(linear['r2'].values)

            self.velocity_genes = (
                self.velocity_genes
                & (self.r2 > self.min_r2)
                & (self.r2 < 0.95)
                & (np.array(linear['coef'].values) > self.min_ratio)
                & (np.max(self.Ms > 0, axis=0) > 0)
                & (np.max(self.Mu > 0, axis=0) > 0)
            )
            print (f'# of velocity genes {self.velocity_genes.sum()} (Criterion: positive regression coefficient between un/spliced counts)')

            lb, ub = np.nanpercentile(self.scaling, [10, 90])
            self.velocity_genes = (
                self.velocity_genes
                & (self.scaling > np.min([lb, 0.03]))
                & (self.scaling < np.max([ub, 3]))
            )
            print (f'# of velocity genes {self.velocity_genes.sum()} (Criterion: std of un/spliced reads should be moderate, w/o extreme values)')

    def fit_curve(self, adata, idx, Ms_scale, Mu_scale, rep=1):
        physical_devices = tf.config.list_physical_devices('GPU')

        if len(physical_devices) == 0 or self.config.GPU == -1:
            tf.config.set_visible_devices([], 'GPU')
            device = '/cpu:0'
            print ('No GPU device has been detected. Switch to CPU mode.')

        else:
            assert self.config.GPU < len(physical_devices), 'Please specify the correct GPU card.'
            tf.config.set_visible_devices(physical_devices[self.config.GPU], 'GPU')

            os.environ["CUDA_VISIBLE_DEVICES"] = f'{self.config.GPU}'
            for gpu in physical_devices:
                tf.config.experimental.set_memory_growth(gpu, True)

            device = f'/gpu:{self.config.GPU}'
            print (f'Using GPU card: {self.config.GPU}')

        with tf.device(device):
            residual, adata = lagrange(
                adata, idx=idx,
                Ms=Ms_scale, Mu=Mu_scale, 
                rep=rep, config=self.config
            )

        return residual, adata

    def fit_velo_genes(self, basis='umap', rep=1):
        idx = self.velocity_genes
        print (f'# of velocity genes {idx.sum()} (Criterion: genes have reads in more than 5% of total cells)')

        if self.config.RESCALE_DATA:
            Ms_scale, Mu_scale = \
                self.Ms, self.Mu / (np.std(self.Mu, axis=0) / np.std(self.Ms, axis=0))
        else:
            Ms_scale, Mu_scale = self.Ms, self.Mu

        assert self.config.GENERAL in ['Deterministic', 'Curve', 'Linear'], \
            'Please specify the correct self.GENERAL in configuration file.'

        if self.config.GENERAL == 'Curve':
            residual, adata = self.fit_curve(self.adata, idx, Ms_scale, Mu_scale, rep=rep)

        if self.config.GENERAL == 'Deterministic':
            residual = self.fit_deterministic(idx, self.Ms, self.Mu, Ms_scale, Mu_scale)

        if self.config.GENERAL == 'Linear':
            self.fit_linear(idx, Ms_scale, Mu_scale, method='kinetic')
            print (np.sum(self.adata[:, idx].var['kinetic_gene'].values) / \
                    np.sum(self.adata[:, idx].var['velocity_genes'].values))
            return self.adata
            
        adata.layers[self.vkey] = residual

        if 'examine_genes' not in adata.uns.keys() and basis != None:
            scv.tl.velocity_graph(adata, sqrt_transform=True)
            scv.tl.velocity_embedding(adata, basis=basis)
            scv.tl.latent_time(adata, min_likelihood=None)
        
        if self.config.FIT_OPTION == '1':
            adata.obs['latent_time'] = adata.obs['latent_time_gm']
            del adata.obs['latent_time_gm']

        return adata


from .velocity import Velocity
import scvelo as scv
import os

def run_model(
    adata,
    label,
    config_file=None,
    normalize=True,
):
    """Preparation and pre-processing function of RNA velocity calculation.
    
    Args:
        adata (str): 
            takes relative of absolute path of Anndata object as input or directly adata object as well.
        label (str): 
            column name in adata.var indicating cell clusters.
        config_file (object): 
            model configuration object, default: None.
        
    Returns:
        adata: 
            modified Anndata object.
    
    """

    from .utils import init_config_summary, init_adata_and_logs
    config, _ = init_config_summary(config=config_file)
    adata, data_path = init_adata_and_logs(adata, config, normalize=normalize)

    scv.settings.presenter_view = True
    scv.settings.verbosity = 0
    scv.settings.file_format_figs = 'png'

    replicates, pre = None, 1e15
    adata_second = adata.copy()

    for rep in range(config.NUM_REP):        
        if rep >= 1:
            adata = adata_second.copy()
            adata.obs['latent_time_gm'] = pre_time_gm

        adata.uns['datapath'] = data_path
        adata.uns['label'] = label
        adata.uns['base_function'] = 'Gaussian'

        if config.BASIS is None:
            basis_keys = ["pca", "tsne", "umap"]
            basis = [key for key in basis_keys if f"X_{key}" in adata.obsm.keys()][-1]
        elif f"X_{config.BASIS}" in adata.obsm.keys():
            basis = config.BASIS
        else:
            raise ValueError('Invalid embedding parameter config.BASIS')
        adata.uns['basis'] = basis

        if 'scNT' in data_path:
            import pandas as pd
            gene_ids = pd.read_csv('../data/scNT/brie_neuron_splicing_time.tsv', delimiter='\t', index_col='GeneID')
            config.VGENES = list(gene_ids.loc[gene_ids['time_FDR'] < 0.01].index)

        model = Velocity(adata, config=config)
        model.get_velo_genes()

        adata = model.fit_velo_genes(basis=basis, rep=rep)
        pre_time_gm = adata.obs['latent_time'].values

        if config.GENERAL != 'Linear':
            replicates = adata if adata.uns['loss'] < pre else replicates
            pre = adata.uns['loss'] if adata.uns['loss'] < pre else pre

    #? change adata to replicates?
    replicates.write(os.path.join(adata.uns['temp'], f'temp_{config.FIT_OPTION}.h5ad'))
    
    if 'examine_genes' in adata.uns.keys():
        from .individual_gene import exam_genes
        exam_genes(adata, adata.uns['examine_genes'])

    return replicates if config.GENERAL != 'Linear' else adata


#%%
#! Base Configuration Class
#! Don't use this class directly. 
#! Instead, sub-class it and override the configurations you need to change.


class Preprocessing(object):
    def __init__(self):
        self.MIN_SHARED_COUNTS = 20

        # (int) # of highly variable genes selected for pre-processing, default 2000
        # consider decreasing to 1500 when # of cells > 10k
        self.N_TOP_GENES = 2000

        self.N_PCS = 30
        self.N_NEIGHBORS = 30

        # (bool) use raw un/spliced counts or first order moments
        self.USE_RAW = False

        # (bool) rescaled Mu/Ms as input based on variance, default True 
        self.RESCALE_DATA = True

class Regularization(object):
    def __init__(self):
        # (bool) regularization on loss function to push peak time away from 0.5
        # mainly used in unified time mode for linear phase portraits
        self.REG_LOSS = True
        # (float) gloablly adjust the magnitude of the penalty, recommend < 0.1
        self.REG_TIMES = 0.075
        # (float) scaling parameter of the regularizer
        self.REG_SCALE = 1

        # (list of tuples) [(gene1, trend1), (gene2, trend2), (gene3, trend3), ...], 
        # a list of genes with trend can be one of {increase, decrease}, default None
        self.GENE_PRIOR = None
        self.GENE_PRIOR_SCALE = 5000

class Optimizer(object):
    def __init__(self):
        # (float) learning rate of the main optimizer
        self.LEARNING_RATE = 1e-2
        # (int) maximum iteration rate of main optimizer
        self.MAX_ITER = 12000

class FittingOption(object):
    def __init__(self):
        # Fitting options under Gaussian model 
        # '1' = Unified-time mode 
        # '2' = Independent mode
        self.FIT_OPTION = '1'

        # (str, experimental) methods to aggregate time metrix, default 'SVD'
        # Max SVD Raw
        self.DENSITY = 'SVD'
        # (str) whether to reorder cell based on relative positions for time assignment
        # Soft_Reorder (default) Hard (for Independent mode)
        self.REORDER_CELL = 'Soft_Reorder'
        # (bool) aggregate gene-specific time to cell time during fitting
        # controlled by self.FIT_OPTION
        self.AGGREGATE_T = True

        # (bool, experimental), whether clip negative predictions to 0, default False
        self.ASSIGN_POS_U = False

        # (bool, experimental) cell time restricted to (0, 1) if False, default False
        self.RESCALE_TIME = False

class VelocityGenes(object):
    def __init__(self):
        # (bool) linear regression $R^2$ on extreme quantile (default) or full data (adjusted)
        # valid when self.VGENES = 'basic'
        self.R2_ADJUST = True

        # (str) selection creteria for velocity genes used in RNA velocity construction, default basic
        # 1. raws, all highly variable genes specified by self.N_TOP_GENES will be used
        # 2. offset, linear regression $R^2$ and coefficient with offset, will override self.R2_ADJUST
        # 3. basic, linear regression $R^2$ and coefficient without offset
        # 4. single gene name, fit this designated gene alone, for model validation purpose only
        # 5. [list of gene names], manually provide a list of genes as velocity genes in string, might improve performance, see scNT
        self.VGENES = 'basic'

        # (float) threshold of R2 at later stage of the optimization proces
        # to capture the dynamics of more genes beside initially selected velocity genes
        # Note: self.AGENES_R2 = 1 will switch to origianl mode with no amplification
        self.AGENES_R2 = 1
        self.AGENES_THRES = 0.61

        # (bool, experimental) exclude cell that have 0 expression in either un/spliced when contributing to loss function
        self.FILTER_CELLS = False

class CellInitialization(object):
    def __init__(self):
        # (str) criteria for cell latent time initialization, default None
        # 1. None, initialized based on the exact order of input expression matrix
        # 2. gcount, str, initialized based on gene counts (https://www.science.org/doi/abs/10.1126/science.aax0249)
        # 3. cluster name, str, use diffusion map based time as initialization
        # 4. [(gene1, trend1), (gene2, trend2), (gene3, trend3), ...], list of tuples, 
        #    a list of genes with trend can be one of {increase, decrease} 
        self.IROOT = None

        # (int) number of random initializations of time points, default 1
        # in rare cases, velocity field generated might be reversed, possibly because stably and monotonically changed genes
        # change this parameter to 2 might do the trick
        self.NUM_REP = 1
        # when self.NUM_REP = 2, the following parameter will determine how the second time will be initialized 
        # re_pre, reverse the inferred cell time of first run
        # re_init, reverse the initialization time of first run
        self.NUM_REP_TIME = 're_pre'

class Configuration():
    def __init__(self):
        Preprocessing.__init__(self)
        Regularization.__init__(self)
        Optimizer.__init__(self)
        FittingOption.__init__(self)
        VelocityGenes.__init__(self)
        CellInitialization.__init__(self)

        # (int) speficy the GPU card for acceleration, default 0
        # -1 will switch to CPU mode
        self.GPU = 0

        # Gaussian Mixture
        self.BASE_FUNCTION = 'Gaussian'

        # Deterministic Curve Linear
        self.GENERAL = 'Curve'

        # (str) embedding format of adata, e.g. pca, tsne, umap, 
        # if None (default), algorithm will choose one automatically
        self.BASIS = None

        # (int, experimental) window size for sliding smoothing of distribution with highest probability
        # useful when self.DENSITY == 'Max'
        # self.WIN_SIZE = 50


#%%
"""
Evaluation utility functions.
This module contains util functions for computing evaluation scores.
"""

import numpy as np
from sklearn.metrics.pairwise import cosine_similarity

def summary_scores(all_scores):
    """Summarize group scores.
    
    Args:
        all_scores (dict{str,list}): 
            {group name: score list of individual cells}.
    
    Returns:
        dict{str,float}: 
            Group-wise aggregation scores.
        float: 
            score aggregated on all samples
        
    """
    sep_scores = {k:np.mean(s) for k, s in all_scores.items() if s}
    overal_agg = np.mean([s for k, s in sep_scores.items() if s])
    return sep_scores, overal_agg

def keep_type(adata, nodes, target, k_cluster):
    """Select cells of targeted type
    
    Args:
        adata (Anndata): 
            Anndata object.
        nodes (list): 
            Indexes for cells
        target (str): 
            Cluster name.
        k_cluster (str): 
            Cluster key in adata.obs dataframe

    Returns:
        list: 
            Selected cells.

    """
    return nodes[adata.obs[k_cluster][nodes].values == target]

def cross_boundary_correctness(
    adata, 
    k_cluster, 
    k_velocity, 
    cluster_edges, 
    return_raw=False, 
    x_emb="X_umap"
):
    """Cross-Boundary Direction Correctness Score (A->B)
    
    Args:
        adata (Anndata): 
            Anndata object.
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        cluster_edges (list of tuples("A", "B")): 
            pairs of clusters has transition direction A->B
        return_raw (bool): 
            return aggregated or raw scores.
        x_emb (str): 
            key to x embedding for visualization.
        
    Returns:
        dict: 
            all_scores indexed by cluster_edges or mean scores indexed by cluster_edges
        float: 
            averaged score over all cells.
        
    """
    scores = {}
    all_scores = {}
    
    x_emb = adata.obsm[x_emb]
    if x_emb == "X_umap":
        v_emb = adata.obsm['{}_umap'.format(k_velocity)]
    else:
        v_emb = adata.obsm[[key for key in adata.obsm if key.startswith(k_velocity)][0]]
        
    for u, v in cluster_edges:
        sel = adata.obs[k_cluster] == u
        nbs = adata.uns['neighbors']['indices'][sel] # [n * 30]
        
        boundary_nodes = map(lambda nodes:keep_type(adata, nodes, v, k_cluster), nbs)
        x_points = x_emb[sel]
        x_velocities = v_emb[sel]
        
        type_score = []
        for x_pos, x_vel, nodes in zip(x_points, x_velocities, boundary_nodes):
            if len(nodes) == 0: continue

            position_dif = x_emb[nodes] - x_pos
            dir_scores = cosine_similarity(position_dif, x_vel.reshape(1,-1)).flatten()
            type_score.append(np.mean(dir_scores))
        
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
        
    if return_raw:
        return all_scores 
    
    return scores, np.mean([sc for sc in scores.values()])

def inner_cluster_coh(adata, k_cluster, k_velocity, return_raw=False):
    """In-cluster Coherence Score.
    
    Args:
        adata (Anndata): 
            Anndata object.
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        return_raw (bool): 
            return aggregated or raw scores.
        
    Returns:
        dict: 
            all_scores indexed by cluster_edges mean scores indexed by cluster_edges
        float: 
            averaged score over all cells.
        
    """
    clusters = np.unique(adata.obs[k_cluster])
    scores = {}
    all_scores = {}

    for cat in clusters:
        sel = adata.obs[k_cluster] == cat
        nbs = adata.uns['neighbors']['indices'][sel]
        same_cat_nodes = map(lambda nodes:keep_type(adata, nodes, cat, k_cluster), nbs)

        velocities = adata.layers[k_velocity]
        cat_vels = velocities[sel]
        cat_score = [cosine_similarity(cat_vels[[ith]], velocities[nodes]).mean() 
                     for ith, nodes in enumerate(same_cat_nodes) 
                     if len(nodes) > 0]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)
    
    if return_raw:
        return all_scores
    
    return scores, np.mean([sc for sc in scores.values()])

def evaluate(
    adata, 
    cluster_edges, 
    k_cluster, 
    k_velocity="velocity", 
    x_emb="X_umap", 
    verbose=True
):
    """Evaluate velocity estimation results using 5 metrics.
    
    Args:
        adata (Anndata): 
            Anndata object.
        cluster_edges (list of tuples("A", "B")): 
            pairs of clusters has transition direction A->B
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        x_emb (str): 
            key to x embedding for visualization.
        
    Returns:
        dict: 
            aggregated metric scores.
    
    """

    from .eval_utils import cross_boundary_correctness
    from .eval_utils import inner_cluster_coh
    crs_bdr_crc = cross_boundary_correctness(adata, k_cluster, k_velocity, cluster_edges, True, x_emb)
    ic_coh = inner_cluster_coh(adata, k_cluster, k_velocity, True)
    
    if verbose:
        print("# Cross-Boundary Direction Correctness (A->B)\n{}\nTotal Mean: {}".format(*summary_scores(crs_bdr_crc)))
        print("# In-cluster Coherence\n{}\nTotal Mean: {}".format(*summary_scores(ic_coh)))
    
    return {
        "Cross-Boundary Direction Correctness (A->B)": crs_bdr_crc,
        "In-cluster Coherence": ic_coh,
    }


import tensorflow as tf
import numpy as np
from scvelo.tools.utils import make_unique_list
from tqdm import tqdm
import scvelo as scv
from .optimize_utils import Model_Utils, exp_args
from .utils import save_vars, new_adata_col, min_max
from .pl import plot_loss

exp = tf.math.exp
log = tf.math.log
sum = tf.math.reduce_sum
mean = tf.math.reduce_mean
sqrt = tf.math.sqrt
abs = tf.math.abs
square = tf.math.square
pow = tf.math.pow
std = tf.math.reduce_std
var = tf.math.reduce_variance

class Recover_Paras(Model_Utils):
    def __init__(
        self,
        adata,
        Ms,
        Mu,
        var_names,
        idx=None,
        rep=1,
        config=None
    ):
        super().__init__(
            adata=adata, 
            var_names=var_names,
            Ms=Ms,
            Mu=Mu,
            config = config
        )

        self.idx = idx
        self.rep = rep
        self.scaling = adata.var['scaling'].values
        self.flag = True

        self.init_pars()
        self.init_vars()
        self.init_weights()
        self.adata.uns['par_names'] = self.default_pars_names
        self.t_cell = self.compute_cell_time(args=None)
        self.pi = tf.constant(np.pi, dtype=tf.float32)

    def compute_cell_time(self, args=None, iter=None, show=False):
        if args != None:
            boundary = (args[4] - 3 * (1 / sqrt(2 * exp(args[3]))), 
                        args[4] + 3 * (1 / sqrt(2 * exp(args[3]))))
            t_range = boundary if self.config.RESCALE_TIME else (0, 1)
            x = self.init_time(t_range, (3000, self.adata.n_vars))

            s_predict, u_predict = self.get_s_u(args, x)
            s_predict = tf.expand_dims(s_predict, axis=0) # 1 3000 d
            u_predict = tf.expand_dims(u_predict, axis=0)
            Mu = tf.expand_dims(self.Mu, axis=1) # n 1 d
            Ms = tf.expand_dims(self.Ms, axis=1)

            t_cell = self.match_time(Ms, Mu, s_predict, u_predict, x.numpy(), iter)

            if self.config.AGGREGATE_T:
                t_cell = tf.reshape(t_cell, (-1, 1))
                t_cell = tf.broadcast_to(t_cell, self.adata.shape)
            
            # plot_phase_portrait(self.adata, args, Ms, Mu, s_predict, u_predict)

        else:
            boundary = (self.t - 3 * (1 / sqrt(2 * exp(self.log_a))), 
                        self.t + 3 * (1 / sqrt(2 * exp(self.log_a))))

            t_cell = self.init_time((0, 1), self.adata.shape)

            if self.rep == 1:
                if self.config.NUM_REP_TIME == 're_init':
                    t_cell = 1 - t_cell
                if self.config.NUM_REP_TIME == 're_pre':
                    t_cell = 1 - self.adata.obs['latent_time_gm'].values
                    t_cell = tf.broadcast_to(t_cell.reshape(-1, 1), self.adata.shape)
            
            if self.rep > 1:
                tf.random.set_seed(np.ceil((self.rep - 1) / 2))
                shuffle = tf.random.shuffle(t_cell)
                t_cell = shuffle if self.rep % 2 == 1 else 1 - shuffle

            if self.config.IROOT == 'gcount':
                print ('---> Use Gene Counts as initial.')
                self.adata.obs['gcount'] = np.sum(self.adata.X.todense() > 0, axis=1)
                g_time = 1 - min_max(self.adata.obs.groupby(self.adata.uns['label'])['gcount'].mean())

                for id in list(g_time.index):
                    self.adata.obs.loc[self.adata.obs[self.adata.uns['label']] == id, 'gcount'] = g_time[id]

                scv.pl.scatter(self.adata, color='gcount', cmap='gnuplot', dpi=100)
                t_cell = tf.cast(
                    tf.broadcast_to(
                        self.adata.obs['gcount'].values.reshape(-1, 1), 
                        self.adata.shape), 
                    tf.float32)

            elif type(self.config.IROOT) == list:
                t_cell, perc = [], []
                for prior in self.config.IROOT:
                    expr = np.array(self.adata[:, prior[0]].layers['Ms'])

                    perc.append(np.max(expr) * 0.75) # modify 0.75 for parameter tuning
                    t_cell.append(min_max(expr) if prior[1] == 'increase' else 1 - min_max(expr))
                
                perc_total = np.sum(perc)
                perc = [perc[i] / perc_total for i in range(len(perc))]
                print (f'assigned weights of IROOT {list(np.around(np.array(perc), 2))}')
                t_cell = [perc[i] * t_cell[i] for i in range(len(perc))]
                t_cell = tf.cast(tf.broadcast_to(np.sum(t_cell, axis=0).reshape(-1, 1), self.adata.shape), tf.float32)

            elif self.config.IROOT in self.adata.obs[self.adata.uns['label']].values:
                print ('Use diffusion pseudotime as initial.')
                import scanpy as sc
                sc.tl.diffmap(self.adata)
                self.adata.uns['iroot'] = \
                    np.flatnonzero(
                        self.adata.obs[self.adata.uns['label']] == self.config.IROOT
                    )[0]
                sc.tl.dpt(self.adata)

                if show:
                    scv.pl.scatter(self.adata, color='dpt_pseudotime', cmap='gnuplot', dpi=100)
                
                t_cell = tf.cast(
                    tf.broadcast_to(
                        self.adata.obs['dpt_pseudotime'].values.reshape(-1, 1), 
                        self.adata.shape), 
                    tf.float32)
        
            else:
                pass

        return t_cell

    def get_loglikelihood(self, distx=None, varx=None):
        n = np.clip(len(distx) - len(self.u) * 0.01, 2, None)
        loglik = -1 / 2 / n * np.sum(distx) / varx
        loglik -= 1 / 2 * np.log(2 * np.pi * varx)

    def amplify_gene(self, t_cell, iter):
        from sklearn import linear_model
        from sklearn.metrics import r2_score
        print (f'\nExaming genes which are not initially considered as velocity genes')

        r2 = np.repeat(-1., self.Ms.shape[1])
        reg = linear_model.LinearRegression()

        for col in tqdm(range(self.Ms.shape[1])):
            if not self.idx[col]:
                y = np.reshape(self.Ms[:, col], (-1, 1))
                x = np.reshape(t_cell, (-1, 1))

                reg.fit(x, y)
                y_pred = reg.predict(x)
                r2[col] = r2_score(y, y_pred)
        
        self.agenes = r2 >= self.config.AGENES_R2
        self.adata.var['amplify_r2'] = r2
        self.adata.var['amplify_genes'] = self.agenes
        self.flag = False

        _ = self.get_log(sum(self.s_r2, axis=0) + sum(self.u_r2, axis=0), True, iter=iter)
        self.used_agenes = np.array(self.adata.var['amplify_genes'].values)
        self.total_genes = self.idx | self.used_agenes

        print (f'# of amplified genes {self.agenes.sum()}, # of used {self.used_agenes.sum()}')
        print (f'# of (velocity + used) genes {self.total_genes.sum()}')

        self.infi_genes = ~np.logical_xor(~self.agenes, self.used_agenes)
        self.adata.var['amplify_infi'] = self.infi_genes
        print (f'# of infinite (or nan) genes {self.infi_genes.sum()}')

    def compute_loss(self, args, t_cell, Ms, Mu, iter, progress_bar):
        self.s_func, self.u_func = self.get_s_u(args, t_cell)
        udiff, sdiff = Mu - self.u_func, Ms - self.s_func

        if (self.config.AGENES_R2 < 1) & (iter > self.agenes_thres):
            self.u_r2 = square(udiff)
            self.s_r2 = square(sdiff) 

            if self.flag:
                self.amplify_gene(t_cell.numpy()[:, 0], iter=iter)

            if iter > int(0.9 * self.config.MAX_ITER) & self.config.REG_LOSS:
                self.s_r2 = self.s_r2 + \
                    std(Ms, axis=0) * self.config.REG_TIMES * \
                    exp(-square(args[4] - 0.5) / self.config.REG_SCALE)

            #compute variance, equivalent to np.var(np.sign(sdiff) * np.sqrt(distx))
            self.vars = mean(self.s_r2, axis=0) \
                - square(mean(tf.math.sign(sdiff) * sqrt(self.s_r2), axis=0))
            self.varu = mean(self.u_r2 * square(self.scaling), axis=0) \
                - square(mean(tf.math.sign(udiff) * sqrt(self.u_r2) * self.scaling, axis=0))

            #! edge case of mRNAs levels to be the same across all cells
            self.vars += tf.cast(self.vars == 0, tf.float32)
            self.varu += tf.cast(self.varu == 0, tf.float32)

            self.u_log_likeli = \
                - (Mu.shape[0] / 2) * log(2 * self.pi * self.varu) \
                - sum(self.u_r2 * square(self.scaling), axis=0) / (2 * self.varu) 
            self.s_log_likeli = \
                - (Ms.shape[0] / 2) * log(2 * self.pi * self.vars) \
                - sum(self.s_r2, axis=0) / (2 * self.vars) 

            error_1 = np.sum(sum(self.u_r2, axis=0).numpy()[self.total_genes]) / np.sum(self.total_genes)
            error_2 = np.sum(sum(self.s_r2, axis=0).numpy()[self.total_genes]) / np.sum(self.total_genes)
            self.se.append(error_1 + error_2)
            progress_bar.set_description(f'Loss (Total): {self.se[-1]:.3f}, (Spliced): {error_2:.3f}, (Unspliced): {error_1:.3f}')

            return self.get_loss(iter,
                                sum(self.s_r2, axis=0), 
                                sum(self.u_r2, axis=0))

        else:
            self.u_r2 = square(udiff)
            self.s_r2 = square(sdiff)

            if (self.config.FIT_OPTION == '1') & \
                (iter > int(0.9 * self.config.MAX_ITER)) & self.config.REG_LOSS:
                self.s_r2 = self.s_r2 + \
                    std(Ms, axis=0) * self.config.REG_TIMES * \
                    exp(-square(args[4] - 0.5) / self.config.REG_SCALE)

            #! convert for self.varu to account for scaling in pre-processing
            self.vars = mean(self.s_r2, axis=0) \
                - square(mean(tf.math.sign(sdiff) * sqrt(self.s_r2), axis=0))
            self.varu = mean(self.u_r2 * square(self.scaling), axis=0) \
                - square(mean(tf.math.sign(udiff) * sqrt(self.u_r2) * self.scaling, axis=0))

            self.u_log_likeli = \
                - (Mu.shape[0] / 2) * log(2 * self.pi * self.varu) \
                - sum(self.u_r2 * square(self.scaling), axis=0) / (2 * self.varu) 
            self.s_log_likeli = \
                - (Ms.shape[0] / 2) * log(2 * self.pi * self.vars) \
                - sum(self.s_r2, axis=0) / (2 * self.vars) 

            error_1 = np.sum(sum(self.u_r2, axis=0).numpy()[self.idx]) / np.sum(self.idx)
            error_2 = np.sum(sum(self.s_r2, axis=0).numpy()[self.idx]) / np.sum(self.idx)
            self.se.append(error_1 + error_2)
            progress_bar.set_description(f'Loss (Total): {self.se[-1]:.3f}, (Spliced): {error_2:.3f}, (Unspliced): {error_1:.3f}')
            
            self.vgene_loss = self.se[-1]
            return self.get_loss(iter,
                                sum(self.s_r2, axis=0), 
                                sum(self.u_r2, axis=0))

    def fit_likelihood(self):
        Ms, Mu, t_cell = self.Ms, self.Mu, self.t_cell
        log_gamma, log_beta, offset = self.log_gamma, self.log_beta, self.offset
        intercept = self.intercept
        log_a, t, log_h = self.log_a, self.t, self.log_h

        from packaging import version
        if version.parse(tf.__version__) >= version.parse('2.11.0'):
            optimizer = tf.keras.optimizers.legacy.Adam(learning_rate=self.init_lr, amsgrad=True)
        else:
            optimizer = tf.keras.optimizers.Adam(learning_rate=self.init_lr, amsgrad=True)
        
        pre = tf.repeat(1e6, Ms.shape[1]) # (2000, )
        self.se, self.m_args, self.m_ur2, self.m_sr2 = [], None, None, None
        self.m_ullf, self.m_sllf = None, None

        progress_bar = tqdm(range(self.config.MAX_ITER))
        for iter in progress_bar:
            with tf.GradientTape() as tape:                
                args = [
                    log_gamma, 
                    log_beta, 
                    offset, 
                    log_a, 
                    t, 
                    log_h, 
                    intercept
                ]
                obj = self.compute_loss(args, t_cell, Ms, Mu, iter, progress_bar)

            stop_cond = self.get_stop_cond(iter, pre, obj)

            if iter > self.agenes_thres + 1:
                self.m_args = self.get_optimal_res(args, self.m_args)
                self.m_ur2 = self.get_optimal_res(self.u_r2, self.m_ur2)
                self.m_sr2 = self.get_optimal_res(self.s_r2, self.m_sr2)
                self.m_ullf = self.get_optimal_res(self.u_log_likeli, self.m_ullf)
                self.m_sllf = self.get_optimal_res(self.s_log_likeli, self.m_sllf)

            if (iter > self.agenes_thres) & \
                (iter == self.config.MAX_ITER - 1 or \
                tf.math.reduce_all(stop_cond) == True or \
                min(self.se[self.agenes_thres + 1:]) * 1.1 < self.se[-1] 
                    if (iter > self.agenes_thres + 1) else False):

                if (iter > int(0.9 * self.config.MAX_ITER)) & self.config.REG_LOSS & \
                    (min(self.se[self.agenes_thres:]) * 1.1 >= self.se[-1]):
                    self.m_args = args
                    self.m_ur2 = self.u_r2
                    self.m_sr2 = self.s_r2
                    self.m_ullf = self.u_log_likeli
                    self.m_sllf = self.s_log_likeli

                t_cell = self.compute_cell_time(args=self.m_args, iter=iter)
                _ = self.get_fit_s(self.m_args, t_cell)
                s_derivative = self.get_s_deri(self.m_args, t_cell)
                # s_derivative = exp(args[1]) * Mu - exp(args[0]) * Ms

                self.post_utils(iter, self.m_args)
                break

            args_to_optimize = self.get_opt_args(iter, args)
            gradients = tape.gradient(target=obj, sources=args_to_optimize)

            # convert gradients of variables with unused genes to 0
            # keep other gradients by multiplying 1
            if (self.config.AGENES_R2 < 1) & (iter > self.agenes_thres):
                convert = tf.cast(self.total_genes, tf.float32)
                processed_grads = [g * convert for g in gradients]
            else:
                convert = tf.cast(self.idx, tf.float32)
                processed_grads = [g * convert for g in gradients]

            optimizer.apply_gradients(zip(processed_grads, args_to_optimize))
            pre = obj

            if iter > 0 and int(iter % 800) == 0:
                t_cell = self.compute_cell_time(args=args, iter=iter)

        self.adata.layers['fit_t'] = t_cell.numpy() if self.config.AGGREGATE_T else t_cell
        self.adata.var['velocity_genes'] = self.total_genes if not self.flag else self.idx
        self.adata.layers['fit_t'][:, ~self.adata.var['velocity_genes'].values] = np.nan

        return self.get_interim_t(t_cell, self.adata.var['velocity_genes'].values), s_derivative.numpy(), self.adata

    def get_optimal_res(self, current, opt):
        return current if min(self.se[self.agenes_thres + 1:]) == self.se[-1] else opt

    def post_utils(self, iter, args):
        # Reshape un/spliced variance to (ngenes, ) and save
        self.adata.var['fit_vars'] = np.squeeze(self.vars)
        self.adata.var['fit_varu'] = np.squeeze(self.varu)

        # Save predicted parameters of RBF kernel to adata
        self.save_pars([item.numpy() for item in args])
        self.adata.var['fit_beta'] /= self.scaling
        self.adata.var['fit_intercept'] *= self.scaling

        # Plotting function for examining model loss
        plot_loss(iter, self.se, self.agenes_thres)

        # Save observations, predictinos and variables locally
        save_vars(self.adata, args, 
                self.s_func.numpy(), self.u_func.numpy(), 
                self.K, self.scaling)

        #! Model loss, log likelihood and BIC based on unspliced counts
        gene_loss = sum(self.m_ur2, axis=0) / self.nobs \
            if self.config.FILTER_CELLS \
            else sum(self.m_ur2, axis=0) / self.Ms.shape[0]

        list_name = ['fit_loss', 'fit_llf']
        list_data = [gene_loss.numpy(), self.m_ullf.numpy()]
        new_adata_col(self.adata, list_name, list_data)

        # Mimimum loss during optimization, might not be the actual minimum
        r2_spliced = 1 - sum(self.m_sr2, axis=0) / var(self.Ms, axis=0) \
            / (self.adata.shape[0] - 1)
        r2_unspliced = 1 - sum(self.m_ur2, axis=0) / var(self.Mu, axis=0) \
            / (self.adata.shape[0] - 1)
        new_adata_col(self.adata, ['fit_sr2', 'fit_ur2'], [r2_spliced.numpy(), r2_unspliced.numpy()])

        tloss = min(self.se[self.agenes_thres + 1:])
        self.adata.uns['loss'] = self.vgene_loss
        print (f'Total loss {tloss:.3f}, vgene loss {self.vgene_loss:.3f}')

    def save_pars(self, paras):
        columns = exp_args(self.adata, 1)
        for i, name in enumerate(self.default_pars_names):
            self.adata.var[f"fit_{name}"] = np.transpose(np.squeeze(paras[i]))

            if name in columns:
                self.adata.var[f"fit_{name}"] = np.exp(self.adata.var[f"fit_{name}"])            

def lagrange(
    adata,
    idx=None,
    Ms=None,
    Mu=None,
    var_names="velocity_genes",
    rep=1,
    config=None
):
    if len(set(adata.var_names)) != len(adata.var_names):
        adata.var_names_make_unique()

    var_names = adata.var_names[idx]
    var_names = make_unique_list(var_names, allow_array=True)

    model = Recover_Paras(
        adata,
        Ms,
        Mu,
        var_names,
        idx=idx,
        rep=rep,
        config=config
    )

    latent_time_gm, s_derivative, adata = model.fit_likelihood()

    if 'latent_time' in adata.obs.columns:
        del adata.obs['latent_time']
    adata.obs['latent_time_gm'] = min_max(latent_time_gm[:, 0])

    return s_derivative, adata


#%%
import os
from time import gmtime, strftime
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

try:
    from setuptools_scm import get_version
    __version__ = get_version(root="..", relative_to=__file__)
    del get_version
    
except (LookupError, ImportError):
    try:
        from importlib_metadata import version 
    except:
        from importlib.metadata import version 
    __version__ = version(__name__)
    del version

print (f'(Running UniTVelo {__version__})')
print (strftime("%Y-%m-%d %H:%M:%S", gmtime()))

from .main import run_model
from .config import Configuration
from .eval_utils import evaluate
from .gene_influence import influence
from .utils import choose_mode, subset_adata


#%%
import io
import os
from setuptools import setup, find_packages

# version of cudatoolkit and tensorflow
REQUIRED = [
    'numpy>=1.20',
    'scikit-learn>=0.22',
    'pandas',
    'scipy>=1.4.1',
    'seaborn',
    'matplotlib>=3.3.0',
    'tqdm',
    'scanpy>=1.5',
    'statsmodels',
    'anndata>=0.7.5',
    'scvelo>=0.2.2',
    'IPython',
    'ipykernel',
    'IProgress',
    'ipywidgets',
    'jupyter',
    'tensorflow>=2.4.1'
]

here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='unitvelo',
    version='0.2.5.1',
    # use_scm_version=True,
    # setup_requires=['setuptools_scm'],

    description='Temporally unified RNA velocity inference',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Mingze Gao',
    author_email='gmz1229@connect.hku.hk',
    python_requires='>=3.7.0',
    url='https://github.com/StatBiomed/UniTVelo',
    packages=find_packages(),

    entry_points={
        'console_scripts': ['unitvelo = unitvelo.main:run_model'],
    },

    install_requires=REQUIRED,
    extras_require={'parallel': ['multiprocessing>=3.8']},
    include_package_data=True,
    license='BSD',
    keywords=[
        'RNA velocity',
        'Unified time',
        'Transcriptomics',
        'Kinetic',
        'Trajectory inference'
    ],

    classifiers=[
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Visualization'
    ]
)


#%%
import os 
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from IPython import display

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    import re
    return [atoi(c) for c in re.split(r'(\d+)', text)]

class Validation():
    def __init__(self, adata, time_metric='latent_time') -> None:
        self.adata = adata
        self.time_metric = time_metric

        if 'latent_time' not in adata.obs.columns:
            import scvelo as scv
            scv.tl.latent_time(adata, min_likelihood=None)         

        if len(set(adata.obs[adata.uns['label']])) > 20:
            self.palette = 'viridis'
        else:
            self.palette = 'tab20'

    def init_data(self, adata):
        DIR = adata.uns['temp']
        f = os.listdir(DIR)
        f.sort(key=natural_keys)

        self.mu = pd.read_csv(f'{DIR}/Mu.csv', index_col=0).add_suffix('_u')
        self.ms = pd.read_csv(f'{DIR}/Ms.csv', index_col=0).add_suffix('_s')

        self.fs = pd.DataFrame(index=self.ms.index)
        self.fu = pd.DataFrame(index=self.ms.index)
        self.var = pd.DataFrame(index=[col[:-2] for col in self.ms.columns])

        for i in f:
            if i.startswith('fits'):
                self.fs = self.fs.join(pd.read_csv(f'{DIR}/{i}', index_col=0).add_suffix(f'_fits'))

            if i.startswith('fitu'):
                self.fu = self.fu.join(pd.read_csv(f'{DIR}/{i}', index_col=0).add_suffix(f'_fitu'))

            if i.startswith('fitvar'):
                df = pd.read_csv(f'{DIR}/{i}', index_col=0)
                df.index = df.index.map(str)
                self.var = self.var.join(df)

        self.lt = adata.obs[self.time_metric].values
        self.label = adata.obs[adata.uns['label']].values

    def concat_data(self):        
        self.msmu = self.ms.join(self.mu)
        self.msmu['labels'] = self.label

        self.msfs = self.ms.join(self.fs)
        self.msfs['labels'] = self.label

        self.mufu = self.mu.join(self.fu)
        self.mufu['labels'] = self.label

        self.fsfu = self.fs.join(self.fu)
        self.fsfu['labels'] = self.label

    def inspect_genes(self, gene_name, adata):
        fig, axes = plt.subplots(2, 2, figsize=(14, 12))

        self.plot_mf(adata, self.msfs[gene_name + '_s'], self.msfs[gene_name + f'_fits'], axes[0][0])
        axes[0][0].plot(
            [np.min(self.fs[gene_name + f'_fits']), np.max(self.fs[gene_name + f'_fits'])], 
            [np.min(self.fs[gene_name + f'_fits']), np.max(self.fs[gene_name + f'_fits'])], 
            ls='--', c='red')
        axes[0][0].set_title('x: ms -- y: fits')

        self.plot_mf(adata, self.mufu[gene_name + '_u'], self.mufu[gene_name + f'_fitu'], axes[0][1])
        axes[0][1].plot(
            [np.min(self.fu[gene_name + f'_fitu']), np.max(self.fu[gene_name + f'_fitu'])], 
            [np.min(self.fu[gene_name + f'_fitu']), np.max(self.fu[gene_name + f'_fitu'])], 
            ls='--', c='red')
        axes[0][1].set_title('x: mu -- y: fitu')

        self.plot_mf(adata, self.msmu[gene_name + '_s'], self.msmu[gene_name + '_u'], axes[1][0])
        axes[1][0].set_title('x: ms -- y: mu')

        self.plot_mf(adata, self.fsfu[gene_name + f'_fits'], self.fsfu[gene_name + f'_fitu'], axes[1][1])
        axes[1][1].set_title('x: fits -- y: fitu')

        plt.show()
    
    def spliced_time(self, gene_name):
        fig, axes = plt.subplots(1, 2, figsize=(12, 4))
        self.sns_plot(self.ms, gene_name + '_s', 'time', 'ms', axes[0])
        self.sns_plot(self.fs, gene_name + f'_fits', 'time', 'fits', axes[1])

    def unspliced_time(self, gene_name):
        fig, axes = plt.subplots(1, 2, figsize=(12, 4))
        self.sns_plot(self.mu, gene_name + '_u', 'time', 'mu', axes[0])
        self.sns_plot(self.fu, gene_name + f'_fitu', 'time', 'fitu', axes[1])
        plt.show()

    def sns_plot(self, data, expression, x, y, loc):
        df = pd.DataFrame(data=[self.lt, data[expression].values], index=[x, y]).T
        df['labels'] = self.label
        sns.scatterplot(x=x, y=y, data=df, hue='labels', sizes=1, palette=self.palette, ax=loc)

    def putative_trans_time(self):
        pass

    def vars_trends(self, gene_name, adata):
        par_names = adata.uns['par_names']
        para = pd.DataFrame(index=par_names, columns=['Values'])
        para['Values'] = self.var.loc[gene_name].values
        self.para = para.iloc[:, -1].T.astype(np.float32)

    def plot_mf(self, adata, s, u, ax=None, hue=None):
        data = [np.squeeze(s), np.squeeze(u)]
        df = pd.DataFrame(data=data, index=['spliced', 'unspliced']).T

        if adata.shape[0] == np.squeeze(s).shape[0]:
            df['labels'] = adata.obs[adata.uns['label']].values
            hue = 'labels'

        sns.scatterplot(x='spliced', y='unspliced', data=df, sizes=1, 
            palette=self.palette, ax=ax, hue=hue)

    def func(self, validate=None, t_cell=None):
        s, u = validate.get_s_u(self.para.values, t_cell)
        return s, u

    def plot_range(self, gene_name, adata, ctype=None):
        #! solving the scaling of unspliced problem afterwards
        from .optimize_utils import exp_args
        from .optimize_utils import Model_Utils
        validate = Model_Utils(config=adata.uns['config'])
        self.vars_trends(gene_name, adata)

        columns = exp_args(adata)
        for col in columns:
            self.para[col] = np.log(self.para[col])

        boundary = (
            np.reshape(self.para['t'] - 3 * (1 / np.sqrt(2 * np.exp(self.para['a']))), (1, 1)), 
            np.reshape(self.para['t'] + 3 * (1 / np.sqrt(2 * np.exp(self.para['a']))), (1, 1))
        )
        
        adata = adata[:, gene_name]
        if ctype != None:
            adata = adata[adata.obs[adata.obs[adata.uns['label']] == ctype].index, :]

        spre, upre = self.func(validate, validate.init_time(boundary, (3000, 1)))
        sone, uone = self.func(validate, validate.init_time((0, 1), (3000, 1)))
        sfit, ufit = self.func(validate, adata.obs['latent_time'].values)

        display.clear_output(wait=True)
        fig, axes = plt.subplots(1, 2, figsize=(12, 4))
        self.plot_mf(adata, adata.layers['Ms'], adata.layers['Mu'] / adata.var['scaling'].values, axes[0])
        self.plot_mf(adata, spre, upre, axes[0])
        self.plot_mf(adata, sone, uone, axes[0])
        self.plot_mf(adata, sfit, ufit, axes[1])

        if 'li_coef' in adata.var.columns:
            x = np.linspace(np.min(adata.layers['Ms']), np.max(adata.layers['Ms']), 1000)
            y = x * adata.var['li_coef'].values
            axes[0].plot(x, y, ls='--', c='red')

        plt.show()

    def plot_scv_fit(self, gene_name, adata):
        DIR = adata.uns['temp']
        self.mu = pd.read_csv(f'{DIR}/Mu.csv', index_col=0).add_suffix('_u')
        self.ms = pd.read_csv(f'{DIR}/Ms.csv', index_col=0).add_suffix('_s')

        self.fu = pd.read_csv(f'{DIR}/scvu.csv', index_col=0).add_suffix(f'_fitu')
        self.fs = pd.read_csv(f'{DIR}/scvs.csv', index_col=0).add_suffix(f'_fits')

        self.lt = adata.obs['latent_time'].values
        self.label = adata.obs[adata.uns['label']].values

        self.concat_data()
        self.inspect_genes(gene_name, adata)
        self.spliced_time(gene_name)
        self.unspliced_time(gene_name)

def exam_genes(adata, gene_name=None, time_metric='latent_time'):
    display.clear_output(wait=True)
    from .individual_gene import Validation
    examine = Validation(adata, time_metric=time_metric)
    examine.init_data(adata)
    examine.concat_data()
    examine.inspect_genes(gene_name, adata)
    examine.spliced_time(gene_name)
    examine.unspliced_time(gene_name)
    examine.vars_trends(gene_name, adata)

def exam_scv(data_path, gene_name, basis, label):
    try:
        import scvelo as scv
    except ModuleNotFoundError:
        print ('Install scVelo via `pip install scvelo`')
    
    adata = scv.read(data_path)
    adata.uns['datapath'] = data_path
    adata.uns['label'] = label
    adata.uns['basis'] = basis

    scv.pp.filter_and_normalize(adata, min_shared_counts=20, n_top_genes=2000)
    scv.pp.moments(adata, n_pcs=30, n_neighbors=30)
    scv.tl.recover_dynamics(adata, n_jobs=20)
    scv.tl.velocity(adata, mode='dynamical')
    scv.tl.velocity_graph(adata)
    scv.tl.latent_time(adata)

    if basis != None:
        scv.pl.velocity_embedding_stream(
            adata, basis=basis, color=label,
            legend_loc='far right', dpi=200, 
            title='scVelo dynamical model'
        )

        scv.pl.scatter(adata, color='latent_time', color_map='gnuplot', size=50)

    if gene_name != None:    
        examine = Validation(adata, time_metric='latent_time')
        examine.plot_scv_fit(gene_name, adata)

    return adata


import numpy as np
import pandas as pd
import os
np.random.seed(42)
import logging
import scvelo as scv

def get_cgene_list():
    s_genes_list = \
        ['Mcm5', 'Pcna', 'Tyms', 'Fen1', 'Mcm2', 'Mcm4', 'Rrm1', 'Ung', 'Gins2',
        'Mcm6', 'Cdca7', 'Dtl', 'Prim1', 'Uhrf1', 'Mlf1ip', 'Hells', 'Rfc2',
        'Rpa2', 'Nasp', 'Rad51ap1', 'Gmnn', 'Wdr76', 'Slbp', 'Ccne2', 'Ubr7',
        'Pold3', 'Msh2', 'Atad2', 'Rad51', 'Rrm2', 'Cdc45', 'Cdc6', 'Exo1', 'Tipin',
        'Dscc1', 'Blm', 'Casp8ap2', 'Usp1', 'Clspn', 'Pola1', 'Chaf1b', 'Brip1', 'E2f8']

    g2m_genes_list = \
        ['Hmgb2', 'Cdk1', 'Nusap1', 'Ube2c', 'Birc5', 'Tpx2', 'Top2a', 'Ndc80',
        'Cks2', 'Nuf2', 'Cks1b', 'Mki67', 'Tmpo', 'Cenpf', 'Tacc3', 'Fam64a',
        'Smc4', 'Ccnb2', 'Ckap2l', 'Ckap2', 'Aurkb', 'Bub1', 'Kif11', 'Anp32e',
        'Tubb4b', 'Gtse1', 'Kif20b', 'Hjurp', 'Cdca3', 'Hn1', 'Cdc20', 'Ttk',
        'Cdc25c', 'Kif2c', 'Rangap1', 'Ncapd2', 'Dlgap5', 'Cdca2', 'Cdca8',
        'Ect2', 'Kif23', 'Hmmr', 'Aurka', 'Psrc1', 'Anln', 'Lbr', 'Ckap5',
        'Cenpe', 'Ctcf', 'Nek2', 'G2e3', 'Gas2l3', 'Cbx5', 'Cenpa']

    return s_genes_list, g2m_genes_list

def new_adata_col(adata, var_names, values):
    for i, name in enumerate(var_names):
        # adata.var[name] = np.zeros(adata.n_vars) * np.nan
        adata.var[name] = values[i]

def get_cycle_gene(adata):
    from .utils import get_cgene_list
    s_genes_list, g2m_genes_list = get_cgene_list()

    # var = adata.var.loc[adata.var['velocity_genes'] == True]
    phase_s, phase_g2 = [], []

    for gene in adata.var.index:
        if gene in s_genes_list:
            phase_s.append(gene)
        if gene in g2m_genes_list:
            phase_g2.append(gene)
    
    return phase_s, phase_g2

def col_corrcoef(raw, fit):
    from .utils import min_max

    res = []
    for col in range(raw.shape[1]):
        corr = np.corrcoef(min_max(raw[:, col]), min_max(fit[:, col]))[0][1]
        res.append(corr)
        
    return np.array(res)

def col_spearman(raw, fit):
    from scipy.stats import spearmanr
    from .utils import min_max

    res = []
    for col in range(raw.shape[1]):
        results, _ = spearmanr(min_max(raw[:, col]), min_max(fit[:, col]))
        res.append(results)

    return np.array(res)

def col_spearman_unorm(raw, fit):
    from scipy.stats import spearmanr

    res = []
    for col in range(raw.shape[1]):
        results, _ = spearmanr(raw[:, col], fit[:, col])
        res.append(results)

    return np.array(res)

def gene_level_spearman(adata, iroot=None, n_top_genes=2000, file_name=None):
    """
    Spearman correlation between the expression profiles fitness 
    of the top-down (scVelo Dynamical mode) or 
    bottom-up strategies (UniTVelo Independent mode)
    Could serve as a systematic comparison (e.g. the entire transcriptome) 
    between two methods    
    """
    
    import scvelo as scv
    import scanpy as sc
    import numpy as np
    import pandas as pd
    from .utils import col_spearman
    from scipy.stats import spearmanr

    print ('---> Running scVelo dynamical mode')
    scvdata = scv.read(adata.uns['datapath'])
    scv.pp.filter_and_normalize(scvdata, min_shared_counts=20, n_top_genes=n_top_genes)
    scv.pp.moments(scvdata, n_pcs=30, n_neighbors=30)
    scv.tl.recover_dynamics(scvdata, n_jobs=20)
    scv.tl.velocity(scvdata, mode='dynamical')
    scv.tl.latent_time(scvdata)

    print ('---> Caculating spearman correlation with reference')
    if 'organoids' in adata.uns['datapath']:
        scvdata.var.index = scvdata.var['gene'].values

    index = adata.var.loc[adata.var['velocity_genes'] == True].index.intersection(scvdata.var.loc[scvdata.var['velocity_genes'] == True].index)
    tadata = adata[:, index].layers['fit_t']
    tscv = scvdata[:, index].layers['fit_t']

    if iroot != None:
        print ('---> Generating reference with diffusion pseudotime')
        sc.tl.diffmap(adata)
        adata.uns['iroot'] = np.flatnonzero(adata.obs[adata.uns['label']] == iroot)[0]
        sc.tl.dpt(adata)

        gt = adata.obs['dpt_pseudotime'].values
        gt = np.broadcast_to(np.reshape(gt, (-1, 1)), tadata.shape)
        resadata = col_spearman(gt, tadata)
        resscv = col_spearman(gt, tscv)

        results, _ = spearmanr(min_max(scvdata.obs['latent_time']), min_max(gt))
        print (np.reshape(results, (-1, 1)))

    if file_name != None:
        print ('---> Generating reference pseudotime with Slingshot')
        slingshot = pd.read_csv(f'./slingshot/{file_name}')
        slingshot = slingshot['x'].values

        slingshot = np.broadcast_to(np.reshape(slingshot, (-1, 1)), tadata.shape)
        resadata = col_spearman(slingshot, tadata)
        resscv = col_spearman(slingshot, tscv)

        results, _ = spearmanr(min_max(scvdata.obs['latent_time']), min_max(slingshot[:, 0]))
        print (np.reshape(results, (-1, 1)))

    corr = pd.DataFrame(index=index, columns=['UniTVelo', 'scVelo'])
    corr['UniTVelo'] = np.reshape(resadata, (-1, 1))
    corr['scVelo'] = np.reshape(resscv, (-1, 1))

    return corr['UniTVelo'].values, corr['scVelo'].values, corr

def gene_level_comparison(corr):
    import seaborn as sns
    import matplotlib.pyplot as plt

    sns.boxplot(data=corr)
    plt.title('Comparison of gene-specific time', fontsize=12)
    plt.ylabel('Spearman Correlation', fontsize=12)
    plt.show()
    plt.close()

    sns.scatterplot(x='UniTVelo', y='scVelo', data=corr)
    plt.axline((0, 0), (0.5, 0.5), color='r')
    plt.show()
    plt.close()

def col_mse(raw, fit):
    from sklearn.metrics import mean_squared_error

    res = []
    for col in range(raw.shape[1]):
        results = mean_squared_error(raw[:, col], fit[:, col])
        res.append(results)
    return np.array(res)

def col_minmax(matrix):
    return (matrix - np.min(matrix, axis=0)) \
        / (np.max(matrix, axis=0) - np.min(matrix, axis=0))

def inv_prob(obs, fit):
    temp = np.abs(obs - fit)
    temp = np.log(np.sum(temp, axis=0))
    temp = np.exp(-temp)
    return temp / np.sum(temp)

def remove_dir(data_path, adata):
    import shutil
    dir = os.path.split(data_path)[0]
    filename = os.path.splitext(os.path.basename(data_path))[0]

    NEW_DIR = os.path.join(dir, filename)
    adata.uns['temp'] = NEW_DIR
    
    if os.path.exists(NEW_DIR):
        shutil.rmtree(NEW_DIR)
    os.mkdir(NEW_DIR)

def save_vars(
    adata, 
    args, 
    fits, 
    fitu, 
    K=1,
    scaling=1
):
    from .optimize_utils import exp_args

    s = pd.DataFrame(data=fits, index=adata.obs.index, columns=adata.var.index)
    u = pd.DataFrame(data=fitu, index=adata.obs.index, columns=adata.var.index)
    ms = pd.DataFrame(data=adata.layers['Ms'], index=adata.obs.index, columns=adata.var.index)
    mu = pd.DataFrame(data=adata.layers['Mu'], index=adata.obs.index, columns=adata.var.index)
    s['label'] = adata.obs[adata.uns['label']].values

    if adata.var.index[0].startswith('ENSMUSG'):
        adata.var.index = adata.var['gene']
        adata.var.index.name = 'index' 

    var = pd.DataFrame(data=np.zeros((adata.shape[1], )), index=adata.var.index)
    del var[0]

    pars = []
    for i in range(len(args)):
        if args[i].shape[0] > 1:
            for k in range(K):
                # par = np.zeros(adata.n_vars) * np.nan
                par = args[i][k, :].numpy()
                pars.append(par)
        else:
            # par = np.zeros(adata.n_vars) * np.nan
            par = args[i].numpy()
            pars.append(par)
    
    for i, name in enumerate(adata.uns['par_names']):
        var[name] = np.transpose(pars[i])
    
    columns = exp_args(adata, K=K)
    for col in columns:
        var[col] = np.exp(var[col])
    
    var['beta'] /= scaling
    var['intercept'] *= scaling
    
    NEW_DIR = adata.uns['temp']

    s.to_csv(f'{NEW_DIR}/fits.csv')
    u.to_csv(f'{NEW_DIR}/fitu.csv')
    var.to_csv(f'{NEW_DIR}/fitvar.csv')
    ms.to_csv(f'{NEW_DIR}/Ms.csv')
    mu.to_csv(f'{NEW_DIR}/Mu.csv')

def min_max(data):
    return (data - np.min(data)) / (np.max(data) - np.min(data))

def make_dense(X):
    from scipy.sparse import issparse

    XA = X.A if issparse(X) and X.ndim == 2 else X.A1 if issparse(X) else X
    if XA.ndim == 2:
        XA = XA[0] if XA.shape[0] == 1 else XA[:, 0] if XA.shape[1] == 1 else XA
    return np.array(XA)

def get_weight(x, y=None, perc=95):
    from scipy.sparse import issparse

    xy_norm = np.array(x.A if issparse(x) else x)
    if y is not None:
        if issparse(y):
            y = y.A
        xy_norm = xy_norm / np.clip(np.max(xy_norm, axis=0), 1e-3, None)
        xy_norm += y / np.clip(np.max(y, axis=0), 1e-3, None)

    if isinstance(perc, int):
        weights = xy_norm >= np.percentile(xy_norm, perc, axis=0)
    else:
        lb, ub = np.percentile(xy_norm, perc, axis=0)
        weights = (xy_norm <= lb) | (xy_norm >= ub)

    return weights

def R2(residual, total):
    r2 = np.ones(residual.shape[1]) - \
        np.sum(residual * residual, axis=0) / \
            np.sum(total * total, axis=0)
    r2[np.isnan(r2)] = 0
    return r2

def OLS(x, y):
    mean_x, mean_y = np.mean(x, axis=0), np.mean(y, axis=0)
    numerator = np.sum(x * y - mean_y * x, axis=0)
    denominator = np.sum(x ** 2 - mean_x * x, axis=0)

    coef_ = numerator / denominator
    inter_ = mean_y - coef_ * mean_x
    return coef_, inter_

def get_model_para(adata):
    var = adata.var.loc[adata.var['velocity_genes'] == True]
    var = var[[
        'scaling', 'fit_vars',
        'fit_varu', 'fit_gamma', 'fit_beta', 'fit_offset0', 
        'fit_a0', 'fit_t0', 'fit_h0', 
        'fit_intercept', 'fit_loss', 'fit_bic', 
        'fit_llf', 'fit_sr2', 'fit_ur2'
    ]]

    if 're_transit' in adata.var.columns:
        var[0].extend(['qua_r2', 'rbf_r2', 're_transit'])

    return var

def reverse_transient(adata, time_metric='latent_time'):
    from scipy.optimize import curve_fit
    import numpy as np
    from tqdm.notebook import tqdm

    adata.var['re_transit'] = False
    adata.var['qua_r2'] = -1.
    adata.var['rbf_r2'] = -1.
    sigma_max = np.max(adata.var.loc[adata.var['velocity_genes'] == True]['fit_a0'])
    celltime = adata.obs[time_metric].values
    
    def quadratic(x, a, b, c):
        return a * (x ** 2) + b * x + c

    def rbf(x, h, sigma, tau):
        return h * np.exp(-sigma * (x - tau) * (x - tau))

    for index, row in tqdm(adata.var.iterrows()):
        if row['velocity_genes']:
            spliced = np.squeeze(np.array(adata[:, index].layers['Ms']))
            popt, _ = curve_fit(quadratic, celltime, spliced)

            fitted = quadratic(celltime, popt[0], popt[1], popt[2])
            ss_res = np.sum((spliced - fitted) ** 2)
            ss_tot = np.sum((spliced - np.mean(spliced)) ** 2)
            r2 = 1 - (ss_res / ss_tot)
            adata.var.loc[index, 'qua_r2'] = r2

            try:
                popt_rbf, _ = curve_fit(rbf, celltime, spliced, 
                    maxfev=10000, 
                    bounds=([1e-2, 1e-3, -np.inf], 
                            [np.max(spliced), sigma_max, +np.inf])
                )

                fitted_rbf = rbf(celltime, popt_rbf[0], popt_rbf[1], popt_rbf[2])
                ss_res_rbf = np.sum((spliced - fitted_rbf) ** 2)
                ss_tot_rbf = np.sum((spliced - np.mean(spliced)) ** 2)
                r2_rbf= 1 - (ss_res_rbf / ss_tot_rbf)            
                adata.var.loc[index, 'rbf_r2'] = r2_rbf
            
            except:
                r2_rbf = 0
                adata.var.loc[index, 'rbf_r2'] = r2_rbf

            if r2 - r2_rbf > 0.075:
                adata.var.loc[index, 're_transit'] = True
    
    from .pl import plot_reverse_tran_scatter
    plot_reverse_tran_scatter(adata)

    re_tran_num = adata.var.loc[adata.var['re_transit'] == True].shape[0]
    re_tran_perc = re_tran_num / adata.var.loc[adata.var['velocity_genes'] == True].shape[0]
    logging.info(f'# of genes which are identified as reverse transient {re_tran_num}')
    logging.info(f'percentage of genes which are identified as reverse transient {re_tran_perc}')
    
    return adata

def choose_mode(adata, label=None):
    print ('This function works as a reference only.')
    print ('For less certain scenario, we also suggest users to try both.')
    print ('---> Checking cell cycle scores...')

    from .utils import get_cgene_list
    import scvelo as scv
    scv.pp.filter_and_normalize(adata, min_shared_counts=20, n_top_genes=2000)

    s, g2m = get_cgene_list()
    num_s = len(adata.var.index.intersection(s))
    num_g2m = len(adata.var.index.intersection(g2m))

    print (f'---> Number of S genes {num_s}/{len(s)}')
    print (f'---> Number of G2M genes {num_g2m}/{len(g2m)}')

    if (num_s / len(s) > 0.5) or (num_g2m / len(g2m) > 0.5):
        print ('Independent mode is recommended, consider setting config.FIT_OPTION = 2')

    else:
        print ('# of cycle genes failed to pass thresholds')
        print ('---> Checking sparse cell types...')
        scv.pp.moments(adata, n_pcs=30, n_neighbors=30)
        adata.obs['cid'] = list(range(adata.shape[0]))

        try:
            neighbors = adata.uns['neighbors']['indices']
        except:
            scv.pp.neighbors(adata, n_pcs=30, n_neighbors=30)
            neighbors = adata.uns['neighbors']['indices']

        ctype_perc = []
        ctype = list(set(adata.obs[label].values))

        for type in ctype:
            temp = adata[adata.obs.loc[adata.obs[label] == type].index, :]
            temp_id = temp.obs['cid'].values
            temp_nei = neighbors[temp_id, 1:].flatten()

            temp_nei = [True if nei in temp_id else False for nei in temp_nei]
            ctype_perc.append(np.sum(temp_nei) / len(temp_nei))
        
        if np.sum(np.array(ctype_perc) > 0.95) >= 3:
            print ('More than two sparse cell types have been detected')
            print ('Independent mode is recommended, consider setting config.FIT_OPTION = 2')
        else:
            print ('Unified-time mode is recommended, consider setting config.FIT_OPTION = 1')

def subset_adata(adata, label=None, proportion=0.5, min_cells=50):
    adata.obs['cid'] = list(range(adata.shape[0]))
    ctype = list(set(adata.obs[label].values))

    subset = []

    for type in ctype:
        temp = adata[adata.obs.loc[adata.obs[label] == type].index, :]
        temp_id = temp.obs['cid'].values

        if len(temp_id) <= min_cells:
            subset.extend(list(temp_id))
        elif int(temp.shape[0] * proportion) <= min_cells:
            subset.extend(list(np.random.choice(temp_id, size=min_cells, replace=False)))
        else:
            subset.extend(list(np.random.choice(temp_id, size=int(temp.shape[0] * proportion), replace=False)))
        
    return adata[np.array(subset), :]

def subset_prediction(adata_subset, adata, config=None):
    from .optimize_utils import Model_Utils

    model = Model_Utils(adata, config=config)
    x = model.init_time((0, 1), (3000, adata.n_vars))
    
    adata.var = adata_subset.var.copy()
    adata.uns['basis'] = adata_subset.uns['basis']
    adata.uns['label'] = adata_subset.uns['label']
    adata.uns['par_names'] = adata_subset.uns['par_names']
    adata.uns['base_function'] = adata_subset.uns['base_function']

    model.total_genes = adata.var['velocity_genes'].values
    model.idx = adata.var['velocity_genes'].values
    scaling = adata.var['scaling'].values

    args_shape = (1, adata.n_vars)
    args = [
        np.broadcast_to(np.log(np.array(adata.var['fit_gamma'].values)), args_shape), 
        np.broadcast_to(np.log(np.array(adata.var['fit_beta'].values * scaling)), args_shape), 
        np.broadcast_to(np.array(adata.var['fit_offset0'].values), args_shape), 
        np.broadcast_to(np.log(np.array(adata.var['fit_a0'].values)), args_shape), 
        np.broadcast_to(np.array(adata.var['fit_t0'].values), args_shape), 
        np.broadcast_to(np.log(np.array(adata.var['fit_h0'].values)), args_shape), 
        np.broadcast_to(np.array(adata.var['fit_intercept'].values / scaling), args_shape)
    ]

    s_predict, s_deri_predict, u_predict = \
        model.get_fit_s(args, x), model.get_s_deri(args, x), model.get_fit_u(x)
    s_predict = tf.expand_dims(s_predict, axis=0) # 1 3000 d
    u_predict = tf.expand_dims(u_predict, axis=0)
    Mu = tf.expand_dims(adata.layers['Mu'] / scaling, axis=1) # n 1 d
    Ms = tf.expand_dims(adata.layers['Ms'], axis=1)

    t_cell = model.match_time(Ms, Mu, s_predict, u_predict, x.numpy(), config.MAX_ITER)
    t_cell = np.reshape(t_cell, (-1, 1))
    t_cell = np.broadcast_to(t_cell, adata.shape)

    adata.layers['fit_t'] = t_cell.copy()
    adata.layers['fit_t'][:, ~adata.var['velocity_genes'].values] = np.nan
    
    model.fit_s = model.get_fit_s(args, t_cell).numpy()
    model.s_deri = model.get_s_deri(args, t_cell).numpy()
    model.fit_u = model.get_fit_u(args).numpy()
    adata.layers['velocity'] = model.s_deri
    
    adata.obs['latent_time_gm'] = min_max(np.nanmean(adata.layers['fit_t'], axis=1))
    scv.tl.velocity_graph(adata, sqrt_transform=True)
    scv.tl.velocity_embedding(adata, basis=adata.uns['basis'])
    scv.tl.latent_time(adata, min_likelihood=None)
        
    if config.FIT_OPTION == '1':
        adata.obs['latent_time'] = adata.obs['latent_time_gm']
        del adata.obs['latent_time_gm']
    
    if os.path.exists(os.path.join(adata_subset.uns['temp'], 'prediction')):
        pass
    else: os.mkdir(os.path.join(adata_subset.uns['temp'], 'prediction'))
    adata.uns['temp'] = os.path.join(adata_subset.uns['temp'], 'prediction')

    import shutil
    shutil.copyfile(os.path.join(adata_subset.uns['temp'], 'fitvar.csv'), 
                    os.path.join(adata.uns['temp'], 'fitvar.csv'))
    
    import pandas as pd
    s = pd.DataFrame(data=model.fit_s, index=adata.obs.index, columns=adata.var.index)
    u = pd.DataFrame(data=model.fit_u, index=adata.obs.index, columns=adata.var.index)
    ms = pd.DataFrame(data=adata.layers['Ms'], index=adata.obs.index, columns=adata.var.index)
    mu = pd.DataFrame(data=adata.layers['Mu'], index=adata.obs.index, columns=adata.var.index)
    s['label'] = adata.obs[adata.uns['label']].values

    NEW_DIR = adata.uns['temp']
    s.to_csv(f'{NEW_DIR}/fits.csv')
    u.to_csv(f'{NEW_DIR}/fitu.csv')
    ms.to_csv(f'{NEW_DIR}/Ms.csv')
    mu.to_csv(f'{NEW_DIR}/Mu.csv')
    adata.write(os.path.join(NEW_DIR, f'predict_adata.h5ad'))

    return adata

def prior_trend_valid(adata, gene_list=None, name='IROOT'):
    import sys
    vgenes_temp = []
    vgenes = adata.var.loc[adata.var['velocity_genes'] == True].index

    file_path = os.path.join(adata.uns['temp'], 'vgenes.txt')
    with open(file_path, 'w') as fp:
        for item in vgenes:
            fp.write("%s\n" % item)

    for prior in gene_list:
        if prior[0] not in vgenes:
            print (f'{prior[0]} of {name} has not been identified as a velocity gene')
        else:
            vgenes_temp.append(prior + (list(adata.var.index).index(prior[0]), ))
    
    if len(vgenes_temp) == 0:
        print (f'---> No genes has been identified as velocity genes')
        print (f'Consider selecting one from {file_path}')
        sys.exit()
    else:
        print (f'Modified {name} {vgenes_temp} with index')
        return vgenes_temp

def init_config_summary(config=None):
    from .config import Configuration
    if config == None:
        print (f'Model configuration file not specified. Default settings with unified-time mode will be used.')
        config = Configuration()

    if config.FIT_OPTION == '1':
        config.DENSITY = 'SVD' if config.GENE_PRIOR == None else 'Raw'
        config.REORDER_CELL = 'Soft_Reorder'
        config.AGGREGATE_T = True

    elif config.FIT_OPTION == '2':
        config.DENSITY = 'Raw'
        config.REORDER_CELL = 'Hard'
        config.AGGREGATE_T = False

    else:
        raise ValueError('config.FIT_OPTION is invalid')

    print ('------> Manully Specified Parameters <------')
    config_ref = Configuration()
    dict_input, dict_ref = vars(config), vars(config_ref)

    para_used = []
    for parameter in dict_ref:
        if dict_input[parameter] != dict_ref[parameter]:
            print (parameter, dict_input[parameter], sep=f':\t')
            para_used.append(parameter)

    print ('------> Model Configuration Settings <------')
    default_para = ['N_TOP_GENES', 
                    'LEARNING_RATE', 
                    'FIT_OPTION', 
                    'DENSITY', 
                    'REORDER_CELL', 
                    'AGGREGATE_T', 
                    'R2_ADJUST', 
                    'GENE_PRIOR', 
                    'VGENES', 
                    'IROOT']

    for parameter in default_para:
        if parameter not in para_used:
            print (parameter, dict_ref[parameter], sep=f':\t')
    
    print ('--------------------------------------------')
    print ('')
    return config, para_used

def init_adata_and_logs(adata, config, normalize=True):
    if type(adata) == str:
        data_path = adata
        adata = scv.read(data_path)

    else:
        cwd = os.getcwd()
        if os.path.exists(os.path.join(cwd, 'res')):
            pass
        else: os.mkdir(os.path.join(cwd, 'res'))

        print (f'Current working dir is {cwd}.')
        print (f'Results will be stored in res folder')
        data_path = os.path.join(cwd, 'res', 'temp.h5ad')
    
    from .utils import remove_dir
    remove_dir(data_path, adata)
    logging.basicConfig(filename=os.path.join(adata.uns['temp'], 'logging.txt'),
                        filemode='a',
                        format='%(asctime)s, %(levelname)s, %(message)s',
                        datefmt='%H:%M:%S',
                        level=logging.INFO)

    if normalize:
        scv.pp.filter_and_normalize(adata, 
                                    min_shared_counts=config.MIN_SHARED_COUNTS, 
                                    n_top_genes=config.N_TOP_GENES)
        print (f"Extracted {adata.var[adata.var['highly_variable'] == True].shape[0]} highly variable genes.")

        print (f'Computing moments for {len(adata.var)} genes with n_neighbors: {config.N_NEIGHBORS} and n_pcs: {config.N_PCS}')
        scv.pp.moments(adata, 
                        n_pcs=config.N_PCS, 
                        n_neighbors=config.N_NEIGHBORS)
    else:
        scv.pp.neighbors(adata)

    print ('')
    return adata, data_path